#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import traceback
from os import PathLike
from pathlib import Path
from typing import Optional, Union

import h5py
import pandas as pd
from anndata import AnnData, read_mtx
from matplotlib.pyplot import imread
from .spatial import set_geometry, to_points
from .utils import get_scale


def read_img_data(path: Union[Path, PathLike], adata: AnnData, res: str = "high") -> AnnData:
    path = Path(path)

    loc = "hires" if res == "high" else "lowres"
    spatial_path = path / "spatial"
    scalefactors_path = spatial_path / "scalefactors_json.json"
    img_path = spatial_path / f"tissue_{loc}_image.png"

    if img_path.exists() and scalefactors_path.exists():
        image = imread(str(img_path))
        adata.uns.setdefault("spatial", {}).setdefault("img", {})
        adata.uns["spatial"]["img"][loc] = image
        adata.uns["spatial"]["scale"] = json.load(scalefactors_path.open("rt"))
    else:
        raise ValueError("Cannot read tissue image or scaling file")
    return adata


def _read_10x_h5(path: PathLike, symbol_as_index: bool = False, dtype: str = "float64") -> Optional[AnnData]:
    """
    Parameters
    ----------
    path : String
        String that designates where the h5 file can be found.

    Returns
    -------
    adata : Anndata object
        Returns anndata object with features and barcode information.

    """
    from scipy.sparse import csr_matrix

    try:
        with h5py.File(path, "r") as f:
            for i in f.keys():
                data = f[i]["data"][()]
                barcodes = f[i]["barcodes"][()].astype("str")
                features = f[i]["features"]
                shape = f[i]["shape"][()]
                indices = f[i]["indices"][()]
                indptr = f[i]["indptr"][()]

            # arr = np.zeros((shape[1],shape[0]))
            # cell_nr = indptr[0]
            # cell_ind = 0
            # for i in range(indices.shape[0]):

            #     arr[cell_ind,indices[i]] = data[i]
            #     if(i > indptr[cell_ind+1]):
            #         cell_ind = cell_ind + 1
            # cell_nr = indptr[cell_ind+1]
            cm = csr_matrix((data, indices, indptr), shape=(shape[1], shape[0]), dtype=dtype)
            # df_feat = pd.DataFrame(np.column_stack((features["id"][()].astype("str"),features["feature_type"][()].astype("str"),
            # features["genome"][()].astype("str"))),index=features["name"][()].astype("str"))

            var_names_key = "id"
            gene_name_key = "name"
            secondary_gene_column_name: str = "symbol"
            if symbol_as_index:
                var_names_key, gene_name_key = gene_name_key, var_names_key
                secondary_gene_column_name = "gene_ids"

            adata = AnnData(
                cm,
                obs=dict(obs_names=barcodes),
                var={
                    "var_names": features[var_names_key][()].astype("str"),
                    secondary_gene_column_name: features[gene_name_key][()].astype("str"),
                    "feature_types": features["feature_type"][()].astype("str"),
                    "genome": features["genome"][()].astype("str"),
                },
                dtype=dtype,
            )

        return adata
    except Exception:
        traceback.print_exc()
        return None
    return adata


def read_10x_counts(
    path: PathLike,
    datatype: Optional[str] = None,
    raw: bool = True,
    prefix: Optional[str] = None,
    symbol_as_index: bool = False,
    dtype: str = "float64",
) -> AnnData:
    path = Path(path)
    if not path.exists():
        raise ValueError(f"Reading with path {path!r} failed, ")

    prefix_str = prefix or ""

    raw_qualifier = "raw" if raw else "filtered"
    h5_file_path = path / f"{prefix_str}{raw_qualifier}_feature_bc_matrix.h5"
    mtx_dir_path = path / f"{prefix_str}{raw_qualifier}_feature_bc_matrix"

    adata: Optional[AnnData] = None

    if (datatype is None and h5_file_path.exists()) or datatype == "h5":
        adata = _read_10x_h5(h5_file_path, symbol_as_index=symbol_as_index, dtype=dtype)
    elif (datatype is None and mtx_dir_path.exists()) or datatype == "mtx":
        adata = _read_10x_mtx(mtx_dir_path, symbol_as_index=symbol_as_index, dtype=dtype)

    if adata is None:
        raise ValueError("Invalid datatype for bc_matrix")

    adata.uns["config"] = {
        "var_names": "symbol" if symbol_as_index else "gene_ids",
        "secondary_var_names": "gene_ids" if symbol_as_index else "symbol",
    }

    return adata


def _read_10x_mtx(path: PathLike, symbol_as_index: bool = False, dtype: str = "float64") -> AnnData:
    path = Path(path)
    genes = pd.read_csv(path / "features.tsv.gz", header=None, sep="\t")
    cells = pd.read_csv(path / "barcodes.tsv.gz", header=None, sep="\t")

    data = read_mtx(path / "matrix.mtx.gz", dtype=dtype).T

    varname_pos, geneids_pos = 0, 1
    gene_column_key = "symbol"
    if symbol_as_index:
        varname_pos, geneids_pos = geneids_pos, varname_pos
        gene_column_key = "gene_ids"

    adata = AnnData(
        data.X,
        obs=dict(obs_names=cells[0].to_numpy()),
        var={
            "var_names": genes[varname_pos].to_numpy(),
            gene_column_key: genes[geneids_pos].to_numpy(),
            "feature_types": genes[2].to_numpy(),
        },
        dtype=dtype,
    )

    return adata


def read_10x_visium(
    path: PathLike,
    datatype: Optional[str] = None,
    raw: bool = True,
    prefix: Optional[str] = None,
    symbol_as_index: bool = False,
    dtype: str = "float64",
    res: str = "hires",
) -> AnnData:
    """

    Parameters
    ----------
    path : PathLike
        Path to visium "out" directory.
    datatype : String, optional
        Either h5 or mtx. If h5 the function will load raw_feature_bc_matrix.h5 or
        filtered_feature_bc_matrix.h5 in the path directory.
        If it is mtx then the function will look for the approprate directory for .gz files. The default is None.
    raw : Boolean, optional
        If false loads only barcodes detected in tissue, otherwise loads the raw barcodes. The default is True.
    prefix : Str, optional
        If there is a prefix to the data this appends it when loading the files. The default is None.

    Raises
    ------
    ValueError
        If the function cannot find files in designated paths, it raises an error.

    Returns
    -------
    adata : Anndata object
        Complete anndata object with spatial information in adata.uns["spatial"] .

    """
    path = Path(path)
    adata = read_10x_counts(path, datatype, raw, prefix, symbol_as_index, dtype=dtype)

    # spatial
    tissue_pos_path = path / "spatial" / "tissue_positions.csv"
    tissue_alt_path = tissue_pos_path.with_stem("tissue_positions_list")

    if tissue_pos_path.exists():
        version = 2
        adata.obs = pd.concat([adata.obs, pd.read_csv(tissue_pos_path).set_index(["barcode"])], axis=1)
    elif tissue_alt_path.exists():
        version = 1
        colnames = ["barcode", "in_tissue", "array_row", "array_col", "pxl_row_in_fullres", "pxl_col_in_fullres"]
        adata.obs = pd.concat([adata.obs, pd.read_csv(tissue_pos_path, header=None, names=colnames).set_index(["barcode"])], axis=1)
    else:
        raise ValueError("Cannot read file tissue_positions.csv")

    adata = read_img_data(path, adata, res=res)
    metadata = {
        "data_source": "visium",
        "img_res": [res for res in ("lowres", "hires") if (path / "spatial" / f"tissue_{res}_image.png").exists()],
        "version": version,
    }
    adata.uns["metadata"] = metadata
    return adata



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Union,
)

import numpy as np
import scipy.sparse as sp
from anndata import AnnData


def is_highres(adata: AnnData) -> bool:
    if "hires" in adata.uns["spatial"]["img"]:
        return True
    if "lowres" in adata.uns["spatial"]["img"]:
        return False
    raise ValueError("Cannot find image data in .uns['spatial']")


def is_lowres(adata: AnnData) -> bool:
    if "lowres" in adata.uns["spatial"]["img"]:
        return True
    if "hires" in adata.uns["spatial"]["img"]:
        return False
    raise ValueError("Cannot find image data in .uns['spatial']")


def make_unique(items: List) -> List:
    items = items[:]
    for i in range(len(items) - 1, -1, -1):
        if items.count(items[i]) > 1:
            items.pop(i)
    return items


def get_scale(adata: AnnData, res: Optional[str] = None) -> float:
    if res not in [None, "hi", "hires", "lo", "lowres"]:
        raise ValueError(f"Unrecognized value {res} for res.")

    scale_key = None
    if is_lowres(adata) and res in [None, "lowres", "lo"]:
        scale_key = "tissue_lowres_scalef"
    if is_highres(adata) and res in [None, "hires", "hi"]:
        scale_key = "tissue_hires_scalef"

    if scale_key is None:
        raise ValueError("Invalid resolution. Make sure the correct image is loaded.")

    return adata.uns["spatial"]["scale"][scale_key]


def add_per_gene_qcmetrics(adata: AnnData, subsets: Dict[str, np.ndarray], force: bool = False) -> None:
    if "sum" not in adata.var.keys() or force:
        adata.var["sum"] = adata.X.sum(axis=0).T  # type: ignore

    if "detected" not in adata.var.keys() or force:
        adata.var["detected"] = np.diff(adata.X.tocsc().indptr)  # type: ignore

    for key, subset in subsets.items():
        sum_key = f"subsets_{key}_sum"
        detected_key = f"subsets_{key}_detected"
        percent_key = f"subsets_{key}_percent"

        ss = adata[subset, :].X

        if sum_key not in adata.var.keys() or force:
            adata.var[sum_key] = ss.sum(axis=0).T  # type: ignore

        if detected_key not in adata.var.keys() or force:
            adata.var[detected_key] = np.diff(ss.tocsc().indptr)  # type: ignore

        if percent_key not in adata.var.keys() or force:
            adata.var[percent_key] = adata.var[sum_key] / adata.var["sum"] * 100


def add_per_cell_qcmetrics(adata: AnnData, subsets: Dict[str, np.ndarray], force: bool = False) -> None:
    if "sum" not in adata.obs.keys() or force:
        adata.obs["sum"] = adata.X.sum(axis=1)  # type: ignore

    if "detected" not in adata.obs.keys() or force:
        adata.obs["detected"] = np.diff(adata.X.tocsr().indptr)  # type: ignore

    for key, subset in subsets.items():
        sum_key = f"subsets_{key}_sum"
        detected_key = f"subsets_{key}_detected"
        percent_key = f"subsets_{key}_percent"

        subset_X = adata[:, subset].X
        if sum_key not in adata.obs.keys() or force:
            adata.obs[sum_key] = subset_X.sum(axis=1)  # type: ignore

        if detected_key not in adata.obs.keys() or force:
            adata.obs[detected_key] = np.diff(subset_X.tocsr().indptr)  # type: ignore

        if percent_key not in adata.obs.keys() or force:
            adata.obs[percent_key] = adata.obs[sum_key] / adata.obs["sum"] * 100


def log_norm_counts(
    adata: AnnData,
    layer: Optional[str] = None,
    inplace: bool = True,
    base: Optional[int] = 2,
    pseudocount: int = 1,
):
    # Roughly equivalent to:
    # target_sum = adata.X.sum(axis=1).mean()
    # sc.pp.normalize_total(adata, target_sum=target_sum)
    # sc.pp.log1p(adata, base=2)

    if not inplace:
        adata = adata.copy()

    X = adata.X if layer is None else adata.layers[layer]

    cell_sums = sp.csr_matrix(X.sum(axis=1)).toarray()
    cell_sums /= cell_sums.mean()

    X = sp.csr_matrix(X / cell_sums)

    log_funcs = {2: np.log2, 10: np.log10, None: np.log}
    other_log = lambda x: np.log(x) / np.log(base)  # type: ignore # noqa: E731
    log = log_funcs.get(base, other_log)

    if pseudocount == 1:
        # This ensures we don't interact with the zeros in X
        X.data = log(X.data + pseudocount)
    else:
        X = log(X + pseudocount)

    if layer is None:
        adata.X = X
    else:
        adata.layers[layer] = X

    return adata


def scale(
    X,
    center=True,
    unit_variance: bool = True,
    center_before_scale: bool = True,
    ddof: int = 1,
):
    is_sparse = isinstance(X, sp.csr_matrix)
    if is_sparse:
        X = X.todense()
    else:
        X = X.copy()

    kwargs = dict(axis=0, keepdims=True)
    if isinstance(X, np.matrix):
        kwargs.pop("keepdims")

    if center and center_before_scale:
        X -= X.mean(**kwargs)

    if unit_variance:
        std = X.std(ddof=ddof, **kwargs)
        w = np.where(std < 1e-8)
        std[w] = 1
        X = np.divide(X, std)

    if center and not center_before_scale:
        X -= X.mean(axis=0)

    return X


def normalize_csr(X: sp.csr_matrix, byrow: bool = True) -> sp.csr_matrix:
    axis = int(byrow)
    sum_ = sp.csr_matrix(X.sum(axis=axis))
    sum_.eliminate_zeros()
    sum_.data = 1 / sum_.data
    sum_ = sp.diags(sum_.toarray().ravel())
    return sum_.dot(X) if byrow else X.dot(sum_)


def kurtosis(x, method: str = "moments"):
    if method != "moments":
        raise NotImplementedError('Only method="moments" is currently implemented')

    n = x.size
    x_bar = x.mean()
    # From asbio::kurt in R:
    # methods of moments kurtosis is
    #   m_4 / m_2^2  with m_j = sum((x-x_mean)**j)/n

    m_2 = np.square(x - x_bar).mean()
    m_4 = np.power(x - x_bar, 4).mean()

    return m_4 / m_2**2


def listify(
    x: Union[None, int, str, Iterable[str], Iterable[int]],
    size: Optional[int] = None,
) -> List[Any]:
    """Converts a string or an iterable of strings to a list of strings.

    Parameters
    ----------
    x : Union[str, Iterable[str]]
        The string or iterable to convert.

    Returns
    -------
    List[str]
        The list of strings.
    """
    nontype = type(None)
    size = size if size is not None else 1
    return [x] * size if isinstance(x, (int, float, str, nontype)) else list(x)



from collections import defaultdict
from typing import (
    Any,
    List,
    Optional,
    Tuple,
    Union,
)

import geopandas as gpd
import networkx as nx
import numpy as np
import pandas as pd
import scipy.sparse as sp
import shapely
from anndata import AnnData
from shapely.geometry import Point, Polygon

from voyagerpy import utils
from voyagerpy.spatial import spatial

# from .spatial import get_default_graph


def compute_weights(coords: gpd.GeoSeries, func="euclidean") -> Any:
    if func == "euclidean":
        point_func = shapely.ops.BaseGeometry.distance  # type: ignore
    else:
        point_func = func

    def row_func(point):
        return coords.apply(point_func, other=point)

    return coords.apply(row_func)


def knn(W: np.ndarray, k: int) -> np.ndarray:
    n = W.shape[0]
    nbrs = np.empty((n, k))
    for i, d in enumerate(W):
        idx = np.argpartition(d, k + 1)[k + 1]
        nbrs[i] = idx[idx != i]
    return nbrs


def dnn(
    W: np.ndarray,
    min_dist: float = 0,
    max_dist: float = 1,
    full: bool = True,
    include_low: bool = False,
    include_high: bool = False,
) -> Union[np.ndarray, List[np.ndarray]]:
    pred00 = lambda d: (min_dist < d) & (d < max_dist)  # noqa E713
    pred10 = lambda d: (min_dist <= d) & (d < max_dist)  # noqa E713
    pred01 = lambda d: (min_dist < d) & (d <= max_dist)  # noqa E713
    pred11 = lambda d: (min_dist <= d) & (d <= max_dist)  # noqa E713

    pred = ((pred00, pred01), (pred10, pred11))[include_low][include_high]

    if full:
        return np.apply_along_axis(pred, 0, W)

    ret = [np.where(pred(d))[0] for d in W]
    return ret


def graph2np(g: Any) -> List[List[Tuple[str, float]]]:
    neighbour_list = []

    return neighbour_list


_spatial_nb_methods = ["tri2nb", "knn", "dnn", "gabriel", "relative", "soi", "poly2nb"]
_dist_types = ["none", "idw", "exp", "dpd"]


def compute_visium_dists(adata, filt):
    select = select_obs(adata, filt)

    coords = adata[select].obs[["array_col", "array_row"]] * np.array([1, np.sqrt(3)])
    positions = gpd.GeoSeries(map(Point, coords.values), index=coords.index)
    dists = compute_weights(positions)
    return dists


def select_obs(adata: AnnData, filt: Optional[Tuple[str, Any]]) -> Union[slice, pd.Series]:
    if filt is None:
        return slice(None)

    filt_key, filt_val = filt
    if not isinstance(filt_val, (tuple, list)):
        filt_val = (filt_val,)

    return adata.obs[filt_key].isin(filt_val)


def compute_visium_graph(adata, filt, dist_key: str) -> sp.csr_matrix:
    if dist_key not in adata.obsp:
        select = select_obs(adata, filt)
        dists = compute_visium_dists(adata, filt)

        if filt is not None:
            (idx,) = np.where(select)
            rows = np.tile(idx, len(idx))
            cols = np.repeat(idx, len(idx))
            matrix_data = (dists.values.ravel(), (rows, cols))

            adata.obsp[dist_key] = sp.csr_matrix(matrix_data, shape=(adata.n_obs, adata.n_obs))
        else:
            adata.obsp[dist_key] = sp.csr_matrix(dists.values, shape=(adata.n_obs, adata.n_obs))

    visium_graph = dnn(adata.obsp[dist_key].todense(), max_dist=3, full=True).astype(float)
    return sp.csr_matrix(visium_graph, shape=(adata.n_obs, adata.n_obs))


def find_visium_graph(
    adata: AnnData, filt=("in_tissue", 1), graph_key: str = "visium", dist_key: str = "dists", row_normalize: bool = True
) -> nx.Graph:
    if graph_key not in adata.obsp:
        adata.obsp[graph_key] = compute_visium_graph(adata, filt, dist_key)
        # adata.obsp[graph_key].eliminate_zeros()

    if row_normalize is True:
        adata.obsp[graph_key] = utils.normalize_csr(adata.obsp[graph_key], byrow=True)

    G = nx.Graph()
    nodes_out, nodes_in = np.where(adata.obsp[graph_key].todense())

    # Add node names or mappings
    edges = list(zip(nodes_out, nodes_in))
    G.add_edges_from(edges)
    nx.relabel_nodes(G, dict(enumerate(adata.obs.index)), copy=False)

    positions: pd.DataFrame = spatial.get_spot_coords(adata, tissue=True, as_df=True)
    positions = dict(zip(positions.index, positions.values))
    nx.set_node_attributes(G, positions, "pos")
    return G



from matplotlib.patches import Rectangle


def boxs(adata, x="cluster", y="in_tissue", color_by: str = None, cmap: str = "dittoseq"):
    assert y is not None

    group_columns = [y] if x is None else [x, y]
    legend = color_by in group_columns
    tab = adata.obs.groupby(group_columns).size()

    rel = tab / tab.max()
    tab = pd.DataFrame({"counts_": tab}, dtype="int")
    tab["density_"] = rel
    size_const = 0.45
    tab["x"] = size_const
    tab["y"] = size_const

    if x is not None:
        levels = tab.index.levels
        x_levels, y_levels = levels
    else:
        x_levels, y_levels = [0], tab.index.categories
    x_size, y_size = len(x_levels), len(y_levels)

    if y_size == 1 and x_size != 1:
        tab["y"] *= tab["density_"]
    elif y_size != 1 and x_size == 1:
        tab["x"] *= tab["density_"]
    else:
        tab["x"] *= np.sqrt(tab["density_"])
        tab["y"] *= np.sqrt(tab["density_"])

    fig, ax = plt.subplots()
    cmap = plt.get_cmap(cmap)

    if x is not None:
        enumerator = [((x_pos, y_pos), (x_key, y_key)) for x_pos, x_key in enumerate(x_levels) for y_pos, y_key in enumerate(y_levels)]
    else:
        enumerator = [((0, y_pos), y_key) for y_pos, y_key in enumerate(y_levels)]

    for (x_pos, y_pos), key in enumerator:
        width, height, n_points = tab.loc[key, ["x", "y", "counts_"]]
        x_offset = x_pos + 0.5 - width
        y_offset = y_pos + 0.5 - height

        points = np.random.random(size=(int(n_points), 2)) * [2 * width, 2 * height] + [x_offset, y_offset]

        color = 0
        if color_by == x:
            color = cmap(x_pos)
        elif color_by == y:
            color = cmap(y_pos)
        else:
            color = cmap(0)

        xy, w, h = (x_offset, y_offset), 2 * width, 2 * height
        rect = Rectangle(xy, w, h, edgecolor="k", facecolor="lightgray", alpha=0.5)
        ax.add_patch(rect)

        label = x_pos if color_by == x else y_pos
        ax.scatter(*points.T, alpha=0.5, s=8, color=color, label=label)

    ax.set_xticks(np.arange(x_size) + 0.5, labels=x_levels)
    ax.set_yticks(np.arange(y_size) + 0.5, labels=y_levels)
    ax.set_xlabel(x)
    ax.set_ylabel(y)
    ax.grid(False)

    if legend:
        handles = ax.collections

        if color_by == x:
            handles = handles[::y_size]
            labels = list(x_levels)
        elif color_by == y:
            handles = handles[:y_size]
            labels = list(y_levels)
        else:
            return ax
        ax.legend(labels=labels, handles=handles, loc="best")

    return ax


x = None
x = "cluster"
ax = boxs(adata, x=x, cmap="tab10", color_by="in_tissue")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 25 14:08:42 2022

@author: sinant
"""

import functools
from copy import deepcopy
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import geopandas as gpd
import numpy as np
import numpy.typing as npt
from anndata import AnnData
from matplotlib import cm, colors, gridspec
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.collections import LineCollection
from matplotlib.colorbar import Colorbar
from matplotlib.figure import Figure
from matplotlib.path import Path as mplPath
from mpl_toolkits.axes_grid1 import make_axes_locatable
from pandas import DataFrame, Series, options
from pandas.api.types import is_categorical_dtype
from scipy import sparse as sp
from scipy.stats import gaussian_kde, linregress

from voyagerpy import spatial, utils

from .cmap_helper import DivergentNorm

# options.mode.chained_assignment = None  # default='warn'

plt.style.use("ggplot")
plt.rcParams["axes.edgecolor"] = "#00000050"
plt.rcParams["axes.facecolor"] = "white"
plt.rcParams["axes.grid.which"] = "both"
plt.rcParams["axes.grid.axis"] = "both"
plt.rcParams["axes.labelsize"] = 9
plt.rcParams["axes.titlesize"] = 11
plt.rcParams["figure.labelsize"] = 11
plt.rcParams["figure.titlesize"] = 13
plt.rcParams["ytick.labelsize"] = 9
plt.rcParams["xtick.labelsize"] = 9
plt.rcParams["grid.alpha"] = 0.1
plt.rcParams["grid.color"] = "k"
plt.rcParams["image.origin"] = "lower"
plt.rcParams["lines.markersize"] = 4
# plt.rcParams["figure.dpi"] = 100  # 100 is the default
# figure.constrained_layout.wspace 0.02
# figure.subplot.wspace 0.2
""" #TODO: 

ax title font_size 10
fig.supylabel fontsize 10 (figure.labelsize)
"""


def imshow(
    adata: AnnData,
    res: Optional[str] = None,
    ax: Optional[Axes] = None,
    tmp: bool = False,
    title: Optional[str] = None,
    **kwargs,
) -> Axes:
    """Show an image stored in adata.uns["spatial"].

    Parameters
    ----------
    adata : AnnData
        The AnnData object containing the image.
    res : Optional[str], optional
        The resolution of the image to show. If None, the first resolution found is used, by default None.
    ax : Optional[Axes], optional
        The axis showing the image. If None, a new axis is created, by default None.
    tmp : bool, optional
        If True, a temporary image is shown, by default False. This is useful to show an image with
        unapplied transformations.
    title : Optional[str], optional
        The title for the axis, by default None

    **kwargs: Parameters passed to matplotlib.axes.Axes.imshow.
    Returns
    -------
    Axes
        The axis showing the image.
    """

    img_key = "img_tmp" if tmp else "img"
    if res is None:
        res = list(adata.uns["spatial"][img_key].keys())[0]

    img = adata.uns["spatial"][img_key][res]

    if ax is None:
        _, axs = plt.subplots()
    else:
        axs = ax

    im_kwargs = dict(origin="lower")
    extent = kwargs.get("extent", None)

    axs.set_xticks([])
    axs.set_yticks([])
    im_kwargs.update(kwargs)
    if title is not None:
        axs.set_title(title)
    if extent is not None:
        left, right, top, bottom = extent
        y_rows, x_cols = img.shape[:2]
        extent = (
            max(0, left),
            min(x_cols, right),
            max(0, top),
            min(y_rows, bottom),
        )
        im_kwargs["extent"] = extent
        axs.imshow(img[top : bottom + 1, left : right + 1], **im_kwargs)
    else:
        axs.imshow(img, **im_kwargs)
    return axs


def configure_violins(
    violins,
    cmap: Optional[str] = None,
    edgecolor: str = "#00000050",
    alpha: float = 0.7,
    facecolor: Optional[str] = None,
) -> Tuple[str, float]:
    colormap = plt.get_cmap(cmap) if isinstance(cmap, (str, type(None))) else cmap
    for i_color, violin in enumerate(violins["bodies"]):
        violin.set_facecolor(facecolor or colormap(i_color))
        violin.set_edgecolor(edgecolor)
        violin.set_alpha(alpha)
    return edgecolor, alpha


def simple_violinplot(
    ax: Axes,
    df: DataFrame,
    y: Union[str, int],
    cmap: Optional[str] = None,
    scatter_points: bool = True,
    jitter: bool = False,
    **kwargs,
):
    violin_opts = dict(showmeans=False, showextrema=False, showmedians=False)
    kwargs.pop("legend", False)

    labels = {
        "x": kwargs.pop("x_label", None),
        "y": kwargs.pop("y_label", y),
    }

    violin_opts.update(kwargs)
    violins = ax.violinplot(df[y], **violin_opts)
    configure_violins(violins, cmap)

    if scatter_points:
        cols = df[y]
        x_vals = np.ones_like(cols)
        x_colors = np.ones_like(cols)
        if cols.ndim == 2:
            x_offsets = np.arange(cols.shape[1]).ravel()
            x_vals += x_offsets
            x_colors += x_offsets
            for violin, x_offset, y_col in zip(
                violins["bodies"], x_offsets, cols.values.T
            ):
                x_vals[:, x_offset] = jitter_points(
                    violin.get_paths()[0],
                    y_col,
                    x=x_offset,
                    jitter=jitter,
                )
        else:
            x_vals = jitter_points(
                violins["bodies"][0].get_paths()[0],
                y_vals=cols.values,
                jitter=jitter,
            )
        y_vals = cols.values

        scatter(
            x_vals,
            y_vals,
            color_by=x_colors,
            cmap=cmap,
            ax=ax,
            is_categorical=True,
            legend=False,
            alpha=0.7,
            labels=labels,
        )
    else:
        ax.set_ylabel(labels["y"])
        ax.set_xlabel(labels["x"])
    ax.set_xticks([])

    return ax


def jitter_points(
    violin: mplPath, y_vals: np.ndarray, x: int = 0, jitter: bool = False
):
    x_vert, y_vert = np.array(violin.cleaned().vertices).T
    at_max_y = np.argmax(y_vert) + 1
    x_vals = x + np.ones_like(y_vals, dtype=float)
    interpolated = np.interp(y_vals, y_vert[:at_max_y], x_vert[:at_max_y])

    if jitter:
        rand = np.random.uniform(size=x_vals.size) * 2 - 1
        # rand = np.linspace(0, 1, x_vals.size) * 2 - 1
        diff = x_vals - interpolated
        x_vals += diff * rand

    return x_vals


def grouped_violinplot(
    ax: Axes,
    df: DataFrame,
    x: str,
    y: str,
    cmap: Optional[str] = None,
    legend: bool = True,
    vert: bool = True,
    x_label: Optional[str] = None,
    y_label: Optional[str] = None,
    scatter_points: bool = True,
    jitter: bool = False,
    **kwargs,
):
    if not vert:
        x, y = y, x

    groups = df.groupby(x)[y].groups
    keys = sorted(groups.keys())
    labels = keys

    if df[x].dtype == "category" and keys == [0, 1]:
        labels = [False, True]

    grouped_data = [df.loc[groups[key], y] for key in keys]
    violin_opts = dict(
        showmeans=False,
        showextrema=False,
        showmedians=False,
        vert=vert,
        widths=0.8,
    )
    violins = ax.violinplot(grouped_data, **violin_opts)
    facecolor = "white" if scatter_points and jitter else None
    configure_violins(violins, cmap, facecolor=facecolor)

    set_ticks = ax.set_xticks if vert else ax.set_yticks
    set_ticks(np.arange(len(keys)) + 1, labels=labels)

    if not vert:
        x, y = y, x

    colormap = plt.get_cmap(cmap)
    if scatter_points:
        for i, (label, violin) in enumerate(zip(labels, violins["bodies"])):
            x_dat, y_dat = [], []
            if scatter_points:
                y_dat = grouped_data[i]
                x_dat = jitter_points(
                    violin.get_paths()[0],
                    y_dat,
                    x=i,
                    jitter=jitter,
                )

            scatter(
                x_dat,
                y_dat,
                ax=ax,
                label=label,
                color=colormap(i),
                **kwargs,
            )

    ax.set_xlabel(x_label or x)
    ax.set_ylabel(y_label or y)

    if not legend:
        return ax

    # ax.legend(bbox_to_anchor=(1.04, 0.5), loc="center left", title=x, frameon=False)
    return ax


def plot_single_barcode_data(
    adata: AnnData,
    y: Union[int, str],
    x: Union[int, str, None] = None,
    obsm: Optional[str] = None,
    cmap: Optional[str] = None,
    ax: Optional[Axes] = None,
    legend: bool = False,
    color_by: Optional[str] = None,
    x_label: Optional[str] = None,
    y_label: Optional[str] = None,
    **kwargs,
):
    if obsm is not None:
        obs = adata.obsm[obsm].copy()
        if color_by is not None and color_by not in obs:
            obs[color_by] = adata.obs[color_by].copy()
    else:
        obs = adata.obs
    if ax is None:
        fig, ax = plt.subplots()
    assert ax is not None

    if is_categorical_dtype(obs[y]):
        if x is None:
            raise NotImplementedError('"Rectangule" plots not implemented')
        elif is_categorical_dtype(obs[x]):
            raise NotImplementedError('"Rectangule" plots not implemented')
        else:
            # Create a horizontal plot, so we group by y instead of x
            ax = grouped_violinplot(
                ax,
                obs,
                x,
                y,
                cmap,
                legend=legend,
                x_label=x_label,
                y_label=y_label,
                vert=False,
            )
    else:
        if x is None:
            ax = simple_violinplot(
                ax, obs, y, cmap, x_label=x_label, y_label=y_label, **kwargs
            )

        elif is_categorical_dtype(obs[x]):
            ax = grouped_violinplot(
                ax,
                obs,
                x,
                y,
                cmap,
                legend=legend,
                x_label=x_label,
                y_label=y_label,
                **kwargs,
            )

        else:
            _scatter_kwargs = dict(
                alpha=0.5,
                labels=dict(x=x_label or x, y=y_label or y),
            )
            _scatter_kwargs.update(kwargs)
            ax = scatter(
                x, y, color_by=color_by, cmap=cmap, ax=ax, data=obs, **_scatter_kwargs
            )

    return ax


def configure_subplots(
    nplots: int, ncol: Optional[int] = 2, **kwargs
) -> Tuple[Figure, npt.NDArray[plt.Axes]]:
    ncol = min(ncol or 2, nplots)
    nrow = int(np.ceil(nplots / ncol))

    # TODO: We may need to increase the space if we draw legends, yticks, etc
    hspace = kwargs.pop("hspace", 0.2)
    wspace = kwargs.pop("wspace", None)

    default_figsize = (6, 6)
    plot_kwargs = {
        "figsize": default_figsize,
        "gridspec_kw": {"wspace": wspace, "hspace": hspace},
        "squeeze": False,
    }
    plot_kwargs.update(kwargs)

    # tight layout incompatible with gridspec_kw arguments wspace and hspace
    if plot_kwargs.get("layout", None) == "tight":
        plot_kwargs.pop("gridspec_kw", None)

    fig, axs = plt.subplots(nrow, ncol, **plot_kwargs)
    if not isinstance(axs, np.ndarray):
        axs = np.array([axs])

    assert isinstance(axs, np.ndarray)
    for ax in axs.flat[nplots:]:
        ax.axis("off")

    return fig, axs


def plot_barcode_data(
    adata: AnnData,
    y: Union[str, Sequence[str]],
    x: Optional[str] = None,
    obsm: Optional[str] = None,
    ncol: Optional[int] = None,
    cmap: Union[
        None, str, colors.ListedColormap, colors.LinearSegmentedColormap
    ] = None,
    color_by: Optional[str] = None,
    sharex: Union[None, Literal["none", "all", "row", "col"], bool] = None,
    sharey: Union[None, str, bool] = None,
    x_label: Union[None, str, Sequence[str]] = None,
    y_label: Union[None, str, Sequence[str]] = None,
    rc_context: Optional[Dict[str, Any]] = None,
    ax: Optional[Axes] = None,
    subplot_kwargs: Optional[Dict] = None,
    **kwargs,
):
    x_features = utils.listify(x)
    y_features = utils.listify(y)

    if x_label is None:
        x_labels = x_features[:]
        if obsm is not None:
            x_labels = [f"{lab}_{obsm}" for lab in x_labels]
    else:
        x_labels = utils.listify(x_label)

    if y_label is None:
        y_labels = y_features[:]
        if obsm is not None:
            y_labels = [f"{lab}_{obsm}" for lab in y_labels]
    else:
        y_labels = utils.listify(y_label)

    del x, y, x_label, y_label

    nplots = len(y_features) * len(x_features)

    if ncol is None:
        ncol = len(x_features)

    if sharex is None:
        if len(y_features) and ncol == len(x_features):
            sharex = "col"
        else:
            sharex = False
    if sharey is None:
        if len(y_features) and ncol == len(x_features):
            sharey = "row"
        else:
            sharey = False

    if ax is not None:
        axs_arr = np.array([ax])
        fig = ax.get_figure()
    else:
        _subplot_kwargs = dict(
            sharex=sharex,
            sharey=sharey,
            figsize=kwargs.pop("figsize", None),
            layout=kwargs.pop("layout", "constrained"),
        )
        default_rc_context = {
            "axes.spines.bottom": True,
            "axes.spines.top": False,
            "axes.spines.left": True,
            "axes.spines.right": False,
            "axes.grid": False,
        }
        _subplot_kwargs.update(subplot_kwargs or {})
        default_rc_context.update(rc_context or {})
        with plt.rc_context(default_rc_context):
            fig, axs_arr = configure_subplots(nplots, ncol, **_subplot_kwargs)

    feature_iterator = (
        (
            (y_lab, y_feat),
            (x_lab, x_feat),
        )
        for (y_lab, y_feat) in zip(y_labels, y_features)
        for (x_lab, x_feat) in zip(x_labels, x_features)
    )

    for i_plot, (y_feat, x_feat) in enumerate(feature_iterator):
        i_col = i_plot % ncol
        add_legend = i_col == ncol - 1
        x_label, x_feat = x_feat
        y_label, y_feat = y_feat

        axs_arr.flat[i_plot] = plot_single_barcode_data(
            adata,
            y=y_feat,
            x=x_feat,
            obsm=obsm,
            cmap=cmap,
            ax=axs_arr.flat[i_plot],
            legend=add_legend,
            color_by=color_by,
            x_label=x_label,
            y_label=y_label,
            **kwargs,
        )

    return axs_arr


def plot_bin2d(
    data: Union[AnnData, DataFrame],
    x: str,
    y: str,
    filt: Optional[str] = None,
    subset: Optional[str] = None,
    bins: int = 101,
    name_true: Optional[str] = None,
    name_false: Optional[str] = None,
    hex_plot: bool = False,
    cmap: Optional[str] = "Blues7",
    cmap_true: Optional[str] = "Reds",
    binwidth: Optional[float] = None,
    ax: Optional[Axes] = None,
    **plot_kwargs,
) -> Axes:
    get_dataframe = lambda df: df.obs if x in df.obs and y in df.obs else df.var
    obs = get_dataframe(data) if isinstance(data, AnnData) else data

    _plot_kwargs = dict(
        bins=bins,
        cmap=cmap,
        cmin=1,
    )

    _legend_kwargs = dict(
        y=_plot_kwargs.pop("y_label", y),
        x=_plot_kwargs.pop("x_label", x),
    )

    figsize = plot_kwargs.pop("figsize", None)
    plot_kwargs.update(_plot_kwargs)

    if hex_plot:
        # hexbin has similar arguments as hist2d, but some names are different
        renaming = [
            ("gridsize", "bins", bins),
            ("mincnt", "cmin", 1),
        ]
        for hex_name, hist_name, default in renaming:
            val = plot_kwargs.pop(hist_name, default)
            plot_kwargs.setdefault(hex_name, val)

        plot_kwargs.setdefault("edgecolor", "#8c8c8c")
        plot_kwargs.setdefault("linewidth", 0.2)

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = ax.get_figure()

    plot_fun = ax.hexbin if hex_plot else ax.hist2d
    x_dat = obs[x]
    y_dat = obs[y]
    del x, y

    if filt is None:
        selection = ...
    elif isinstance(filt, str):
        selection = obs[filt].astype(bool)
    else:
        selection = filt

    x_dat = x_dat[selection]
    y_dat = y_dat[selection]

    if subset is None:
        if filt is None:
            selection = ...
        elif isinstance(filt, str):
            selection = obs[filt].astype(bool)
        else:
            selection = filt

        im = plot_fun(x_dat, y_dat, **plot_kwargs)  # type: ignore
        cbar = fig.colorbar(im[-1] if isinstance(im, tuple) else im, label="count")
    else:
        subset_name = ""
        if isinstance(subset, str):
            subset_name = subset
            subset_true = obs[subset].astype(bool)
            subset_false = (1 - subset_true).astype(bool)
        else:
            subset_true = subset.astype(bool)
            subset_false = (1 - subset).astype(bool)

        name_true = name_true or subset_name or "True"
        name_false = name_false or (f"!{subset_name}" if subset_name else "False")

        x_min, x_max = x_dat.min(), x_dat.max()
        y_min, y_max = y_dat.min(), y_dat.max()

        if hex_plot:
            plot_kwargs["extent"] = (x_min, x_max, y_min, y_max)
        else:
            plot_kwargs["range"] = [[x_min, x_max], [y_min, y_max]]

        im_false = plot_fun(x_dat[subset_false], y_dat[subset_false], **plot_kwargs)

        plot_kwargs["cmap"] = cmap_true
        im_true = plot_fun(x_dat[subset_true], y_dat[subset_true], **plot_kwargs)

        plt.colorbar(
            im_false[-1] if isinstance(im_false, tuple) else im_false, label=name_false
        )
        plt.colorbar(
            im_true[-1] if isinstance(im_true, tuple) else im_true, label=name_true
        )

    ax.set_ylabel(_legend_kwargs["y"])
    ax.set_xlabel(_legend_kwargs["x"])

    ax.grid()
    return ax


def plot_expression(
    adata: AnnData,
    gene: Union[str, Sequence[str]],
    y: Optional[str] = None,
    obsm: Optional[str] = None,
    ax: Union[None, Axes, np.ndarray[Axes]] = None,
    **kwargs,
):
    _rc_params = {
        "axes.spines.top": False,
        "axes.spines.right": False,
        "axes.grid": False,
    }
    with plt.rc_context(_rc_params):
        if y is None:
            return plot_expression_violin(adata, gene, **kwargs)

        else:
            return plot_expression_scatter(adata, gene, y, obsm=obsm, ax=ax, **kwargs)


def plot_expression_scatter(
    adata: AnnData,
    gene: str,
    y: str,
    obsm: Optional[str] = None,
    ax: Union[None, Axes] = None,
    color_by: Optional[str] = None,
    layer: Optional[str] = None,
    show_symbol: bool = True,
    **kwargs,
):
    obs = adata.obs if obsm is None else adata.obsm[obsm]
    X = adata.X if layer is None else adata.layers[layer]

    if sp.issparse(X):
        X = X.toarray()
    elif isinstance(X, np.matrix):
        X = np.array(X)

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    color_data = obs if color_by in obs else adata.obs

    if isinstance(gene, str):
        gene_idx = adata.var_names.get_loc(gene)
        x_data = X[:, gene_idx]
    else:
        x_data = gene
    if isinstance(y, str):
        y_data = obs[y]
        y_symbol = adata.var.at[y, "symbol"]
    else:
        y_data = y
        y_symbol = None

    ax = scatter(
        x_data,
        y_data,
        label=gene,
        color_by=color_by,
        data=color_data,
        ax=ax,
        **kwargs,
    )

    ax.set_title(y_symbol if show_symbol else gene)
    ax.set_xlabel("Expression" + f" ({layer})" if layer is not None else "")
    if y_symbol is not None:
        y_lab = y_symbol if show_symbol else gene
        ax.set_ylabel(y_lab + f" {obsm}" if obsm is not None else "")

    return ax


def plot_expression_violin(
    adata: AnnData,
    gene: Union[str, Sequence[str]],
    groupby: Optional[str] = None,
    ncol: Optional[int] = 2,
    show_symbol: bool = False,
    layer: Optional[str] = None,
    cmap: Optional[str] = None,
    subplot_kwargs: Optional[Dict] = None,
    **kwargs,
):
    # genes = [genes] if isinstance(genes, str) else genes[:]
    genes = utils.listify(gene)
    gene_ids = []

    secondary_column = adata.uns["config"]["secondary_var_names"]

    for genes in genes:
        if genes in adata.var[secondary_column].values:
            new_genes = adata.var.index[adata.var[secondary_column] == genes]
            gene_ids.extend(new_genes.tolist())
        else:
            assert genes in adata.var_names
            gene_ids.append(genes)

    X = (adata.X if layer is None else adata.layers[layer]).copy()

    obs = (adata.obs[[groupby]] if groupby is not None else adata.obs).copy()
    for gene_id in gene_ids:
        gene_idx = adata.var_names.get_loc(gene_id)
        counts = X[:, gene_idx]

        if sp.issparse(counts):
            counts = counts.A

        obs[gene_id] = counts

    _subplot_kwargs = dict(
        figsize=kwargs.pop("figsize", None),
        sharey=True,
        sharex=True,
        layout="constrained",
    )
    _subplot_kwargs.update(subplot_kwargs or {})

    nplots = len(gene_ids) if groupby is not None else 1
    fig, axs = configure_subplots(nplots, ncol, **_subplot_kwargs)

    if groupby is None:
        gene_ids = [gene_ids]

    for ax, gene_id in zip(axs.flat, gene_ids):
        if groupby is not None:
            grouped_violinplot(
                ax, obs, groupby, gene_id, legend=False, cmap=cmap, **kwargs
            )
            title = gene_id
            if show_symbol and secondary_column == "symbol":
                title = adata.var.at[gene_id, secondary_column]
            elif show_symbol:
                title = gene_id

            ax.set_title(title)
        else:
            simple_violinplot(ax, obs, gene_id, legend=False, cmap=cmap, **kwargs)
            labels = gene_id
            if show_symbol and secondary_column == "symbol":
                labels = adata.var.loc[gene_id, "symbol"]
            elif show_symbol:
                labels = gene_id
            ax.set_xticks(np.arange(len(gene_id)) + 1, labels=labels, rotation=60)

        ax.set_ylabel("")
        ax.set_xlabel("")

    fig.supylabel("Expression" + f" ({layer})" if layer is not None else "")
    if groupby is not None:
        fig.supxlabel(groupby)

    return axs


def plot_features_bin2d(adata: AnnData, *args, **kwargs) -> Axes:
    return plot_bin2d(adata.var, *args, **kwargs)


def plot_barcodes_bin2d(adata: AnnData, *args, **kwargs) -> Axes:
    return plot_bin2d(adata.obs, *args, **kwargs)


def rcDecorator(context):
    def decorator(func):
        @functools.wraps(func)
        def inner_wrapper(*args, **kwargs):
            with plt.rc_context(context):
                return func(*args, **kwargs)

        return inner_wrapper

    return decorator


def nogrid(func):
    context = {"axes.grid": False}

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with plt.rc_context(context):
            return func(*args, **kwargs)

    return wrapper


def plot_spatial_feature(
    adata: AnnData,
    features: Union[str, Sequence[str]],
    ncol: int = 2,
    barcode_geom: Optional[str] = None,
    annot_geom: Optional[str] = None,
    subset_barcodes: Union[None, slice, Sequence[str]] = None,
    layer: Optional[str] = None,
    obsm: Optional[str] = None,
    geom_style: Optional[Dict] = {},
    annot_style: Optional[Dict] = {},
    alpha: float = 0.2,
    divergent: bool = False,
    color: Optional[str] = None,
    _ax: Union[None, Axes, Iterable[Axes]] = None,
    legend: bool = True,
    cmap: Optional[str] = "Blues7",
    cat_cmap: Optional[str] = None,
    div_cmap: str = "roma",
    subplot_kwargs: Optional[Dict] = None,
    legend_kwargs: Optional[Dict] = None,
    dimension: str = "barcode",
    image_kwargs: Optional[Dict] = None,
    figtitle: Optional[str] = None,
    feature_labels: Union[None, str, Sequence[Optional[str]]] = None,
    rc_context: Optional[Dict[str, Any]] = None,
    show_symbol: bool = True,
    **kwargs,
) -> Union[np.ndarray, Any]:
    feat_ls = utils.listify(features)

    assert_basic_spatial_features(adata, dimension, errors="raise")
    if obsm is not None and obsm not in adata.obsm:
        raise KeyError(f'Key "{obsm}" is not found in adata.obsm')

    # adata = adata.copy()

    secondary_gene_column = adata.uns["config"]["secondary_var_names"]
    added_features = []

    features_to_pop = []

    var_features = []
    labeled_features = []

    df = adata.obs if obsm is None else adata.obsm[obsm]
    df_repr = f"adata.obs" if obsm is None else f'adata.obsm["{obsm}"]'

    feature_labels = (
        feature_labels if feature_labels is not None else [None] * len(feat_ls)
    )
    feature_labels = (
        [feature_labels]
        if not isinstance(feature_labels, (tuple, list))
        else feature_labels[:]
    )

    for feature, label in zip(feat_ls, feature_labels):
        label = label if label is not None else feature
        is_gene = feature in adata.var_names
        if is_gene and show_symbol:
            label = adata.var.at[feature, secondary_gene_column]
        if feature in df:
            labeled_features.append((feature, label))
            continue
        if is_gene:
            labeled_features.append((feature, label))
            var_features.append(feature)
            continue

        # Add the indices of the feature found in the secondary_var_names column
        if feature in adata.var[secondary_gene_column].values:
            ids = adata.var[adata.var[secondary_gene_column] == feature].index
            added_features.extend(ids)
            var_features.extend(ids)

            labeled_features.extend([(idx, feature) for idx in ids])
            features_to_pop.append(feature)
            continue

        raise ValueError(f"Cannot find {feature!r} in {df_repr} or gene names")

    # Rename the features found via the secondary column
    feat_ls.extend(added_features)
    for feature in features_to_pop:
        feat_ls.remove(feature)

    # Remove duplicates
    for i_feature in range(len(feat_ls) - 1, -1, -1):
        feature = feat_ls[i_feature]
        if feat_ls.count(feature) > 1:
            feat_ls.pop(i_feature)

    del feat_ls
    # Select the spots to work with

    barcode_selection = subset_barcodes if subset_barcodes is not None else slice(None)
    gene_selection = (
        slice(None) if not var_features else utils.make_unique(var_features)
    )

    if dimension == "barcode":
        geo = adata.obsm["geometry"].loc[barcode_selection].copy()
        gene_selection = gene_selection if var_features else 0
    elif dimension == "gene":
        geo = adata.varm["geometry"].loc[gene_selection].copy()
    else:
        geo = adata.uns["spatial"]["geometry"][dimension].copy()

    adata = adata[barcode_selection, gene_selection].copy()
    obs = adata.obs if obsm is None else adata.obsm[obsm]

    if var_features:
        if layer is None:
            columns = adata[:, var_features].X.toarray()
        else:
            columns = adata[:, var_features].layers[layer].toarray()

        obs[var_features] = columns

    # check if barcode geometry exists
    if barcode_geom is not None:
        if barcode_geom not in geo:
            raise ValueError(f"Cannot find {barcode_geom!r} data in adata.obs")

        # if barcode_geom is not spot polygons, change the default
        # geometry of the observation matrix, so we can plot it
        if barcode_geom != geo.geometry.name:
            geo.set_geometry(barcode_geom)

    n_features = len(labeled_features)
    ncol = min(ncol, n_features)

    # use a divergent colormap
    if divergent:
        cmap = div_cmap

    _subplot_kwargs = dict(
        layout="constrained",
        figsize=kwargs.pop("figsize", None),
        sharex=True,
        sharey=True,
        squeeze=False,
        hspace=0.2 if legend else 0.05,
        wspace=0.1,
    )
    _subplot_kwargs.update(subplot_kwargs or {})

    # create the subplots with right cols and rows

    _rc_context = {
        "axes.grid": False,
        "figure.frameon": False,
        "axes.spines.bottom": False,
        "axes.spines.top": False,
        "axes.spines.left": False,
        "axes.spines.right": False,
        "xtick.bottom": False,
        "xtick.labelbottom": False,
        "ytick.left": False,
        "ytick.labelleft": False,
    }

    geom_is_poly = geo.geometry.geom_type[0] == "Polygon"

    # This is to match the vignettes.
    # visium_10x and nonspatial don't have the same look for this function.
    if not geom_is_poly:
        for key in _rc_context:
            _rc_context[key] = True

    _rc_context.update(rc_context or {})
    if _ax is None:
        with plt.rc_context(_rc_context):
            fig, axs = configure_subplots(
                nplots=n_features, ncol=ncol, **_subplot_kwargs
            )
        # plt.subplots_adjust(wspace = 1/ncols +  0.2)
    else:
        if isinstance(_ax, Axes):
            axs = np.array([_ax])
        elif not isinstance(_ax, np.ndarray):
            axs = np.array(_ax)
        else:
            axs = _ax

        fig = axs.flat[0].get_figure()

    kwargs.setdefault("s", 4)
    if geom_is_poly:
        kwargs.setdefault("markersize", kwargs.pop("s", None))

    # iterate over features to plot

    for _ax, (feature, label) in zip(axs.flat, labeled_features):
        legend_kwargs_ = deepcopy(legend_kwargs or {})
        _image_kwargs = deepcopy(image_kwargs or {})
        _legend = legend
        curr_cmap = cmap
        values = obs[feature]

        extra_kwargs = dict()
        if values.dtype != "category":
            curr_cmap = cmap
            vmax = None

            legend_kwargs_.setdefault("title", label)
            legend_kwargs_.setdefault("orientation", "vertical")

            if divergent:
                vmin = values.min()
                vmax = values.max()
                vcenter = 0
                extra_kwargs["norm"] = DivergentNorm(vmin, vmax, vcenter)

        else:
            #  colorbar for discrete categories if pandas column is categorical
            curr_cmap = cat_cmap
            vmax = cm.get_cmap(cat_cmap).N
            _legend = False

        if color is not None:
            curr_cmap = None

        if image_kwargs is not None:
            crop_img = _image_kwargs.pop("crop", False)
            extent = None
            if crop_img:
                x = geo.centroid.x
                y = geo.centroid.y
                pad = _image_kwargs.pop("pad", 5)
                x_min, y_min = np.maximum(
                    np.floor([x.min(), y.min()]).astype(int) - pad, 0
                )
                x_max, y_max = np.ceil([x.max(), y.max()]).astype(int) + pad
                extent = (x_min, x_max, y_min, y_max)
            _ax = imshow(adata, None, _ax, extent=extent)

        if geom_is_poly:
            legend_kwargs_.pop("title", None)

            if _legend:
                divider = make_axes_locatable(_ax)
                cax = divider.append_axes("right", size="5%", pad=0.05)
                extra_kwargs["cax"] = cax
                cax.set_title(label)

            _ax = geo.plot(
                column=values,
                ax=_ax,
                color=color,
                legend=_legend,
                cmap=plt.get_cmap(curr_cmap),
                vmax=vmax,
                legend_kwds=legend_kwargs_,
                **extra_kwargs,
                **geom_style,
                **kwargs,
            )

            # This would only happen for categorical data
            if legend and not _legend:
                cmap_colors = plt.get_cmap(curr_cmap).colors
                legend_dict = {
                    lab: color
                    for lab, color in zip(sorted(np.unique(values)), cmap_colors)
                }
                for key, color in legend_dict.items():
                    _ax.scatter([], [], label=key, color=color)
                _ax.legend(
                    loc="center left",
                    bbox_to_anchor=(1.04, 0.5),
                    title=label,
                    frameon=False,
                )

        else:
            _ax = scatter(
                geo.geometry.x,
                geo.geometry.y,
                ax=_ax,
                color_by=feature,
                data=obs,
                legend_kwargs=legend_kwargs_,
                cmap=curr_cmap,
                **extra_kwargs,
                **kwargs,
            )

        if annot_geom is not None:
            if annot_geom in adata.uns["spatial"]["geom"]:
                # check annot_style is dict with correct values
                annot_kwargs = dict(ax=_ax, color="blue", alpha=alpha, **kwargs)
                annot_kwargs.update(annot_style or {})

                plg = adata.uns["spatial"]["geom"][annot_geom]
                gpd.GeoSeries(plg).plot(**annot_kwargs)
            else:
                raise ValueError(
                    f"Cannot find {annot_geom!r} data in adata.uns['spatial']['geom']"
                )

    if figtitle is not None:
        fig.suptitle(figtitle, x=0, ha="left", va="bottom")

    return axs


def assert_basic_spatial_features(
    adata, dimension="barcode", errors: str = "raise"
) -> Tuple[bool, str]:
    ret = True
    errors_to_raise = []
    get_geom_prompt = "Consider run voyagerpy.spatial.get_geom(adata) first."
    set_geom_prompt = "Consider running voyagerpy.spatial.set_geometry first"
    if dimension == "barcode":
        if "geometry" not in adata.obsm:
            error_msg = "geometry dataframe does not exist. " + set_geom_prompt
            if errors == "raise":
                raise KeyError(error_msg)
            return False, error_msg
        try:
            geom_name = adata.obsm["geometry"].geometry.name
        except AttributeError as ex:
            error_msg = ex.args[0]
            if errors == "raise":
                raise ex
            return False, error_msg
    # TODO: add checks for genes and annotgeom
    # TODO: deprecate
    return True, ""

    if not isinstance(adata.obs, gpd.GeoDataFrame):
        error_msg = "adata.obs must be a geopandas.GeoDataFrame. " + get_geom_prompt
        if errors == "raise":
            raise TypeError(error_msg)
        return False, error_msg

    if adata.obs.geometry.name not in adata.obs:
        error_msg = (
            f'"{adata.obs.geometry.name}" must be a column in adata.obs. '
            + get_geom_prompt
        )
        if errors == "raise":
            raise KeyError(error_msg)
        return False, error_msg
    if "geom" not in adata.uns["spatial"]:
        error_msg = '"geom" must be a key in adata.uns["spatial"]. ' + get_geom_prompt
        if errors == "raise":
            raise KeyError(error_msg)
        return False, error_msg

    return True, ""


def plot_local_result(
    adata: AnnData, obsm: str, features: Union[str, Sequence[str]], **kwargs
):
    if obsm not in adata.obsm:
        raise KeyError(f"`{obsm}` not found in adata.obsm.")

    kwargs.setdefault("figtitle", obsm.replace("_", " ").capitalize())

    rc_context = {
        "axes.grid": True,
        "figure.frameon": True,
        "axes.spines.bottom": True,
        "axes.spines.top": True,
        "axes.spines.left": True,
        "axes.spines.right": True,
    }

    axs = plot_spatial_feature(
        adata, features=features, obsm=obsm, rc_context=rc_context, **kwargs
    )
    return axs


def spatial_reduced_dim(
    adata: AnnData,
    dimred: str,
    ncomponents: Union[int, Sequence[int]],
    **kwargs,
):
    adata = adata.copy()
    if isinstance(ncomponents, (list, tuple, range)):
        dims = list(ncomponents)
    elif isinstance(ncomponents, int):
        dims = list(range(ncomponents))
    else:
        raise TypeError("features must be a integer or a sequence of integers")

    # check input
    assert_basic_spatial_features(adata, errors="raise")

    if dimred not in adata.obsm:
        raise ValueError(f"Cannot find {dimred!r} in adata.obsm")

    # create df for dimension reduction
    reductions = adata.obsm[dimred][:, dims]
    feat_names = [f"{dimred}{i}" for i in dims]
    adata.obs[feat_names] = reductions

    axs = plot_spatial_feature(
        adata=adata,
        features=feat_names,
        **kwargs,
    )

    fig = axs.flat[0].get_figure()
    fig.suptitle(dimred, x=0, ha="left", fontsize="xx-large", va="bottom")
    return axs


def add_colorbar_discrete(
    ax,
    fig,
    cmap,
    cbar_title: str,
    cat_names: list,
    scale: bool = False,
    title_kwargs: Optional[Dict] = None,
) -> Colorbar:
    # catnr = adata.obs[feat_ls[i]].unique().shape[0]

    colormap = cm.get_cmap(cmap)
    del cmap

    n_values = len(cat_names)
    n_colors = colormap.N if scale else n_values
    if not scale:
        colormap = colors.ListedColormap(colormap.colors[:n_colors])

    bounds = list(range(n_values + 1))
    n_bounds = cm.get_cmap(colormap).N if scale else n_values + 1
    norm = colors.BoundaryNorm(bounds, n_bounds)

    ticks = [x + 0.5 for x in range(n_values)]

    cbar = fig.colorbar(
        cm.ScalarMappable(cmap=colormap, norm=norm),
        ax=ax,
        extendfrac="auto",
        ticks=ticks,
        spacing="uniform",
        orientation="vertical",
        drawedges=True,
        shrink=0.3,
    )

    title_kwargs = title_kwargs or {}
    cbar.ax.set_title(cbar_title, **title_kwargs)
    cbar.ax.grid(False, "both")
    cbar.ax.set_yticklabels(cat_names)

    cbar.outline.set_visible(False)
    cbar.dividers.set_color("white")
    cbar.dividers.set_linewidth(5)

    return cbar


def subplots_single_colorbar(
    nplot: int = 1,
    ncol: int = 1,
    cax_width: float = 0.2,
    cax_space: float = 0.4,
    **kwargs,
):
    ncol = min(ncol or 2, nplot)
    nrow = int(np.ceil(nplot / ncol))

    if not kwargs.pop("legend", True):
        return *configure_subplots(nplot, ncol, **kwargs), None

    figsize = kwargs.get("figsize", None)

    if isinstance(figsize, tuple):
        figsize = (figsize[0] + cax_width + cax_space, figsize[1])
        kwargs["figsize"] = figsize

    wspace = kwargs.pop("wspace", 0.05)
    hspace = kwargs.pop("hspace", 0.05)

    fig = plt.figure(**kwargs)

    # Configure the gridspec size
    (total_width, _) = fig.get_size_inches()
    plot_width = total_width - cax_width - cax_space

    right = plot_width / total_width
    left = (plot_width + cax_space) / total_width

    has_layout = kwargs.get("layout", None) is not None
    if has_layout:
        ax_width = plot_width / ncol
        width_ratios = [ax_width] * ncol + [cax_width]
        main_spec = fig.add_gridspec(
            nrow, ncol + 1, wspace=wspace, hspace=hspace, width_ratios=width_ratios
        )
        cax_spec = main_spec[:, -1]
    else:
        gridspec_kw = dict(
            wspace=wspace,
            hspace=hspace,
            right=right,
            # left=0.0,
        )
        main_spec = fig.add_gridspec(nrow, ncol, **gridspec_kw)

    axs = np.array(
        [
            [
                fig.add_subplot(
                    main_spec[row, col],
                )
                for col in range(ncol)
            ]
            for row in range(nrow)
        ]
    )

    # Make the last column of the grid the colorbar
    if has_layout:
        cax = fig.add_subplot(main_spec[:, -1])
    else:
        caxspec_kw = dict(
            right=1.0,
            left=left,
        )
        cax_spec = fig.add_gridspec(nrow, 1, **caxspec_kw)
        cax = fig.add_subplot(cax_spec[:, -1])

    cax.set_frame_on(False)
    cax.grid(False)
    cax.set_xticks([])
    cax.set_yticks([])

    for ax in axs.flat[nplot:]:
        ax.axis("off")

    return fig, axs, cax


def plot_dim_loadings(
    adata: AnnData,
    components: Union[int, Sequence[int]],
    ncol: int = 2,
    n_loadings: int = 10,
    show_symbol: bool = True,
    varm: str = "PCs",
    **kwargs,
):
    components = (
        list(range(components)) if isinstance(components, int) else list(components)
    )
    dat = adata.varm[varm][:, components]
    nplots = len(components)

    _subplot_kwargs = dict(
        nplots=nplots,
        ncol=ncol,
        figsize=kwargs.pop("figsize", None),
        sharex=kwargs.pop("sharex", True),
        layout=kwargs.pop("layout", "constrained"),
        hspace=0.15,
    )

    fig, axs = configure_subplots(**_subplot_kwargs)

    show_symbol = show_symbol and adata.uns["config"]["var_names"] != "symbol"

    n_min = n_loadings // 2
    n_max = n_loadings - n_min
    for i, (ax, dim) in enumerate(zip(axs.flat, components)):
        ax.set_title(f"{varm} {dim}")

        # get indices of sorted dim loadings (ascending)
        idx = np.argsort(dat[:, i])
        idx = np.hstack([idx[:n_min], idx[-n_max:]])

        genes = adata.var.index[idx]
        if show_symbol:
            genes = adata.var.loc[genes, "symbol"]

        loadings = dat[idx, i]
        hlines = [((0, row), (loading, row)) for row, loading in enumerate(loadings)]

        line_segments = LineCollection(hlines, colors="k")
        ax.add_collection(line_segments)

        ax.scatter(loadings, genes, c="b")
        ax.axvline(0, linestyle="--", c="k")

        # Show ticks at the bottom of each column
        ax.tick_params("x", labelbottom=i + ncol >= len(components))

    fig.supxlabel("Loading")
    fig.supylabel("Gene")

    return axs


def elbow_plot(
    adata: AnnData, ndims: int = 20, reduction: str = "pca", ax: Optional[Axes] = None
):
    if ax is None:
        fig, ax = plt.subplots()
    var_ratio = adata.uns[reduction]["variance_ratio"][:ndims]
    ndims = var_ratio.size
    ax.plot(
        np.arange(ndims, dtype=int),
        var_ratio * 100,
        marker=".",
        markersize=6,
        linewidth=1,
        c="k",
    )
    ax.set_xticks(np.arange(0, ndims + 1, 3, dtype=int))
    ax.set_ylabel("Variance explained (%)")
    ax.set_xlabel("PC")
    return ax


def plot_pca(
    adata: AnnData,
    ndim: int = 5,
    cmap: Optional[str] = None,
    color_by: str = "cluster",
    obsm="X_pca",
    legend_kwargs: Optional[Dict[str, Any]] = None,
    subplot_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
):
    data = adata.obsm[obsm]
    rc_context = {"axes.edgecolor": "#00000050", "axes.grid.which": "both"}
    _subplot_kwargs = dict(
        figsize=kwargs.pop("figsize", None),
    )
    _subplot_kwargs.update(subplot_kwargs or {})

    with plt.rc_context(rc_context):
        fig, axs, cax = subplots_single_colorbar(ndim * ndim, ndim, **_subplot_kwargs)

    var_expl = np.round(adata.uns["pca"]["variance_ratio"] * 100).astype(int)

    _legend_kwargs = dict(
        bbox_to_anchor=(1.04, 0.5),
        frameon=False,
        loc="center left",
        num=None,
        title=color_by,
    )
    _legend_kwargs.update(legend_kwargs or {})
    scatter_kwargs = dict(
        color_by=color_by,
        data=adata.obs,
        s=8,
        alpha=0.5,
        cmap=cmap,
        vmin=0,
        vmax=plt.get_cmap(cmap).N,
        legend=False,
    )

    scatter_kwargs.update(kwargs)

    for row in range(ndim):
        if ndim > 1:
            scatters_in_row = [ax for i, ax in enumerate(axs[row, :]) if i != row]
            scatters_in_row[0].get_shared_y_axes().join(
                scatters_in_row[0], *scatters_in_row
            )
            axs[1, row].get_shared_x_axes().join(axs[1, row], *axs[1:, row])

        for col in range(ndim):
            ax = axs[row, col]

            if row != col:
                ax = scatter(*data[:, (col, row)].T, ax=ax, **scatter_kwargs)
            if row < ndim - 1:
                ax.set_xticklabels([])
            if col > 0:
                ax.set_yticklabels([])
            ax.set_frame_on(True)

        axs[0, row].set_xlabel(f"PC{row} ({var_expl[row]:d}%)")
        axs[0, row].xaxis.set_label_position("top")
        axs[row, -1].set_ylabel(
            f"PC{row} ({var_expl[row]:d}%)", rotation=270, labelpad=15
        )
        axs[row, -1].yaxis.set_label_position("right")

        density = gaussian_kde(data[:, row])
        xs = np.linspace(data[:, row].min(), data[:, row].max(), 200)
        axs[row, row].plot(xs, density(xs), c="k", linewidth=1)

    if ndim > 1:
        legend_elements = (
            axs.flat[1]
            .collections[0]
            .legend_elements(num=_legend_kwargs.pop("num", None))
        )
        # cax.legend(*legend_elements, loc="center", title=colorby, frameon=False, num=None)
        cax.legend(*legend_elements, **_legend_kwargs)
    else:
        cax.remove()

    return axs


def contour_plot(
    ax: Axes,
    x: Union[str, np.ndarray, Series],
    y: Union[str, np.ndarray, Series],
    data: Any = None,
    shape: Tuple[int, int] = (100, 100),
    levels: int = 7,
    colors: Union[str, Sequence[str]] = "cyan",
    linewidths: Optional[float] = 1,
    origin: Optional[str] = None,
) -> Axes:
    if data is not None:
        xdat = data[x] if isinstance(x, str) else x[:]
        ydat = data[y] if isinstance(y, str) else y[:]
    elif not (isinstance(x, str) or isinstance(y, str)):
        xdat = x
        ydat = y
    else:
        raise ValueError("x and y must both be arrays if data is None")

    points = np.vstack([xdat, ydat])

    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()

    xdim, ydim = shape
    xdim = complex(0, xdim)
    ydim = complex(0, ydim)

    X, Y = np.mgrid[xmin:xmax:xdim, ymin:ymax:ydim]
    positions = np.vstack([X.ravel(), Y.ravel()])
    kernel = gaussian_kde(points)
    Z = np.reshape(kernel(positions), X.shape)

    ax.contour(
        X, Y, Z, levels=levels, colors=colors, linewidths=linewidths, origin=origin
    )

    return ax


def moran_plot(
    adata: AnnData,
    feature: Union[str, Sequence],
    graph_name: Union[None, str, np.ndarray] = None,
    color_by: Optional[str] = None,
    xlim: Optional[Tuple[float, float]] = None,
    ylim: Optional[Tuple[float, float]] = None,
    ax: Optional[Axes] = None,
    contour_kwargs: Optional[Dict[str, Any]] = None,
    show_symbol: bool = True,
    ncol: int = 2,
    legend: bool = True,
    subplot_kwargs: Optional[Dict[str, Any]] = None,
    layer: Optional[str] = None,
    **scatter_kwargs,
) -> Axes:
    adata = adata.copy()

    features = [feature] if isinstance(feature, str) else feature
    del feature

    nplot = len(features)
    ncol = min(ncol, nplot)
    nrow = int(np.ceil(nplot / ncol))

    _subplot_kwargs = dict(
        layout=None if nplot == 1 else "constrained",
        wspace=0.2,
        hspace=0.2,
        cax_space=0.2,
        cax_width=0.01,
        figsize=scatter_kwargs.pop("figsize", None),
        legend=legend,
    )
    _subplot_kwargs.update(subplot_kwargs or {})

    if ax is None:
        fig, axs, cax = subplots_single_colorbar(nplot, ncol, **_subplot_kwargs)
    else:
        if isinstance(ax, Axes):
            axs = np.array([[ax]])
        if axs.size < nplot:
            raise ValueError("Fewer axes than plots.")

    X = adata.X if layer is None else adata.layers[layer].A

    for i_plot, (feature, ax) in enumerate(zip(features, axs.flat)):
        lagged_feature = f"lagged_{feature}"
        if lagged_feature not in adata.obs:
            spatial.compute_spatial_lag(
                adata, feature, graph_name=graph_name, inplace=True
            )

        rc_context = {
            "axes.grid": True,
            "axes.spines.bottom": True,
            "axes.spines.top": True,
            "axes.spines.left": True,
            "axes.spines.right": True,
        }

        y_label = f"Spatially lagged {feature}"

        _contour_kwargs = dict(
            shape=(150, 150),
            levels=7,
            colors="cyan",
            linewidths=1,
        )
        _contour_kwargs.update(contour_kwargs or {})

        labels = dict(x=feature, y=y_label)

        if feature in adata.var_names:
            i_feature = adata.var_names.get_loc(feature)
            adata.obs[feature] = X[:, i_feature]
            if show_symbol:
                symbol = adata.var.at[
                    feature, adata.uns["config"]["secondary_var_names"]
                ]
                labels["x"] = symbol
                labels["y"] = f"Spatially lagged {symbol}"

        ax = scatter(
            x=feature,
            y=lagged_feature,
            color_by=color_by,
            labels=labels,
            rc_context=rc_context,
            fitline_kwargs=dict(color="b"),
            contour_kwargs={"colors": "cyan"},
            data=adata.obs,
            ax=ax,
            legend=legend and (i_plot == nplot - 1),
            legend_kwargs=dict(cax=(cax if i_plot == nplot - 1 else None)),
            **scatter_kwargs,
        )

        if xlim is not None:
            ax.set_xlim(*xlim)
        if ylim is not None:
            ax.set_ylim(*ylim)

        ax.axvline(adata.obs[feature].mean(), linestyle="--", c="k", alpha=0.5)
        ax.axhline(adata.obs[lagged_feature].mean(), linestyle="--", c="k", alpha=0.5)

        ax.set_aspect("equal")

    return axs


def plot_moran_mc(
    adata: AnnData,
    feature: Union[str, Sequence[str]],
    graph_name: Optional[str] = None,
    cmap: Optional[str] = None,
    ax: Optional[Axes] = None,
    legend_title: str = "feature",
    **kwargs,
):
    linewidth = kwargs.pop("linewidth", None)
    _kwargs = dict(figsize=None)
    _kwargs.update(kwargs)
    if ax is None:
        fig, (ax,) = configure_subplots(1, 1, squeeze=True, **_kwargs)

    ax.set_prop_cycle("color", plt.get_cmap(cmap).colors)

    if graph_name is None:
        graph_name = spatial.get_default_graph(adata)
    moran_df = adata.uns["spatial"].get("moran", {}).get(graph_name, None)
    moran_sim_dict = adata.uns["spatial"].get("moran_mc", {}).get(graph_name, None)

    if moran_df is None or moran_sim_dict is None:
        raise KeyError(
            "The AnnData object does not have the results needed for this plot.\
 Please run `voyagerpy.spatial.moran` with permutations > 0 for the selected features."
        )

    features = [feature] if isinstance(feature, str) else feature[:]

    Is = moran_df.loc[feature, "I"]
    minI, maxI = Is.min(), Is.max()

    for feat, I in zip(features, Is):
        sim = moran_sim_dict[feat]["sim"]
        label = adata.var.loc[feat, "symbol"]

        x1 = min(minI, sim.min())
        x2 = max(maxI, sim.max())
        xs = np.linspace(x1, x2, 200)
        kernel = gaussian_kde(sim)
        (p,) = ax.plot(xs, kernel(xs), label=label, linewidth=linewidth)
        ax.axvline(I, color=p.get_c(), linewidth=linewidth)

    ax.legend(
        loc="center left", bbox_to_anchor=(1.04, 0.5), title=legend_title, frameon=False
    )

    return ax


def plot_barcode_histogram(
    adata: AnnData,
    feature: Union[str, Sequence[str]],
    color_by: Optional[str] = None,
    figsize: Optional[Tuple[float, float]] = None,
    ncol: int = 1,
    cmap: Optional[str] = None,
    bins: int = 100,
    log: bool = True,
    stacked: bool = False,
    histtype: str = "step",
    obsm: Optional[str] = None,
    label: Optional[Sequence[str]] = None,
    subplot_kwargs: Optional[Dict[str, Any]] = None,
    **hist_kwargs,
) -> np.ndarray[Axes]:
    features: List[str] = utils.listify(feature)  # type: ignore
    nplot = len(features)
    ncol = min(ncol, nplot)

    if obsm is not None:
        if obsm not in adata.obsm:
            raise KeyError(f'"{obsm}" not found in adata.obsm')
        df: DataFrame = adata.obsm[obsm].copy()  # type: ignore
        if color_by is not None and color_by not in df:
            df[color_by] = adata.obs[color_by].copy()
    else:
        df = adata.obs.copy()

    _subplots_kwargs = dict(figsize=figsize, layout="constrained", hspace=None)
    _subplots_kwargs.update(subplot_kwargs or {})

    labels = [label] if isinstance(label, str) else label
    del label
    if labels is not None and len(labels) != len(features):
        raise ValueError("labels must be the same length as features")

    if color_by is not None:
        fig, axs, cax = subplots_single_colorbar(nplot, ncol, **_subplots_kwargs)
        all_feats = features + [color_by]
        keys, groups = zip(*df[all_feats].groupby(color_by).groups.items())

        colormap = plt.get_cmap(cmap)
        colors = [colormap(i) for i in range(len(keys))]

        for i_ax, (feat, ax) in enumerate(zip(features, axs.flat)):
            hist_range = np.array([df[feat].min(), df[feat].max()])

            hist_data = [
                np.histogram(
                    df.loc[group, feat].values,
                    bins=bins,
                    range=hist_range,
                )
                for group in groups
            ]

            counts, bin_edges = zip(*hist_data)
            bin_edges = np.array(bin_edges)
            counts = np.array(counts)
            if histtype != "line":
                n, bins_, rects = ax.hist(
                    bin_edges[:, :-1].T,
                    bins=bins,
                    log=log,
                    label=keys,
                    stacked=stacked,
                    weights=counts.T,
                    color=colors[: len(keys)],
                    histtype=histtype,
                    **hist_kwargs,
                )
            else:
                ax.set_prop_cycle("color", plt.get_cmap(cmap).colors)

                centers = np.diff(bin_edges) / 2 + bin_edges[:, :-1]
                bin_width = np.diff(bin_edges).mean(axis=1, keepdims=True)

                # CLEANUP: We're adding bins to both ends to make the lines go to zero
                if True:
                    zeros = np.zeros((centers.shape[0], 1))
                    counts = np.hstack([zeros, counts, zeros])
                    centers = np.hstack(
                        [
                            centers[:, :1] - bin_width,
                            centers,
                            centers[:, -1:] + bin_width,
                        ]
                    )

                # we plot counts + log to map zero to zero without losing plotted data
                counts += log
                rects = ax.plot(centers.T, counts.T, **hist_kwargs)

                if log:
                    ax.set_yscale("log")

            ax.grid(False, "minor")
            ax.set_xlabel(feat if labels is None else labels[i_ax])
            ax.xaxis.set_label_position("top")

        if histtype.startswith("step"):
            # Hack to get the actual handles
            handles = [patch[0] for patch in rects]
        else:
            handles = rects
        cax.legend(
            handles=handles,
            labels=keys,
            loc="center left",
            title=color_by,
            frameon=False,
        )
    else:
        fig, axs = configure_subplots(nplot, ncol)
        # TODO

    fig.supylabel("count")
    fig.supxlabel("values")
    if obsm is not None:
        fig.suptitle(f"Barcode histogram for {obsm}", ha="left", x=0.0)
    return axs


def plot_correlogram(
    adata,
    graph_name=None,
    metric="moran",
    order=None,
    show_symbol: bool = True,
    ax=None,
    features=None,
    cmap=None,
):
    if ax is None:
        fig, ax = plt.subplots()
    ax.set_prop_cycle("color", plt.get_cmap(cmap).colors)
    if graph_name is None:
        graph_name = spatial.get_default_graph(adata)

    corr_dict = (
        adata.uns.setdefault("spatial", {})
        .setdefault(metric, {})
        .setdefault("correlogram", {})
    )
    if graph_name not in corr_dict:
        raise KeyError(
            f"Graph {graph_name} not found in adata.uns['spatial']['{metric}']['correlogram']"
        )

    df = corr_dict[graph_name]
    # df = adata.uns["spatial"][metric]["correlogram"][graph_name]
    if order is None:
        order = slice(None)
    if isinstance(order, tuple):
        order = slice(*order)
    if features is None:
        features = df.index
    cols = df.columns[order]

    for feature in features:
        ss = df.loc[feature, cols]
        label = adata.var.at[feature, "symbol"] if show_symbol else feature
        ax.plot(ss.T, "-o", label=label)
    ax.legend(bbox_to_anchor=(1.04, 0.5), loc="center left", title="feature")
    ax.axhline(0, linestyle="--")
    ax.set_ylim(-0.03, 1.03)
    ax.set_ylabel(metric.title())
    ax.set_xlabel("Lag")
    return ax


def plot_features_histogram(
    adata: AnnData,
    features: Union[str, Sequence[str]],
    fill_by: Optional[str] = None,
    figsize: Optional[Tuple[float, float]] = None,
    ncol: int = 1,
    cmap: Optional[str] = None,
    bins: int = 100,
    log: bool = True,
    stacked: bool = False,
    histtype: str = "step",
    markers: Union[None, str, Sequence[str]] = None,
    show_symbol: bool = True,
    **hist_kwargs,
) -> np.ndarray[Axes]:
    features = [features] if isinstance(features, str) else features
    nplot = len(features)
    ncol = min(ncol, nplot)
    nrow = int(np.ceil(nplot / ncol))

    _hist_kwargs = dict(edgecolor="w")

    _hist_kwargs.update(hist_kwargs)

    if fill_by is not None:
        fig, axs, cax = subplots_single_colorbar(nrow, ncol, figsize=figsize)
        all_feats = features + [fill_by]
        keys, groups = zip(*adata.obs[all_feats].groupby(fill_by).groups.items())

        colormap = plt.get_cmap(cmap)
        colors = [colormap(i) for i in range(len(keys))]

        for feat, ax in zip(features, axs.flat):
            # these should probably be .var
            hist_range = np.array([adata.obs[feat].min(), adata.obs[feat].max()])
            hist_mid = hist_range.mean()
            hist_range = tuple((hist_range - hist_mid) * 1.05 + hist_mid)

            hist_data = [
                np.histogram(
                    adata.obs.loc[group, feat].values,
                    bins=bins,
                    range=hist_range,
                    **_hist_kwargs,
                )
                for group in groups
            ]

            counts, bin_edges = zip(*hist_data)
            bin_edges = np.array(bin_edges)
            counts = np.array(counts)
            if histtype != "line":
                n, bins_, rects = ax.hist(
                    bin_edges[:, :-1].T,
                    bins=bins,
                    range=hist_range,
                    log=log,
                    label=keys,
                    stacked=stacked,
                    weights=counts.T,
                    color=colors[: len(keys)],
                    histtype=histtype,
                    **_hist_kwargs,
                )
            else:
                # TODO: make this match the barcode histogram
                ax.set_prop_cycle("color", plt.get_cmap(cmap).colors)

                centers = np.diff(bin_edges) / 2 + bin_edges[:, :-1]
                rects = ax.plot(centers.T, np.maximum(0.2, counts.T))

                if log:
                    ax.set_yscale("log")

                ax.set_ylim(0.2, None)

            ax.grid(False, "minor")
            ax.set_xlabel(feat, size=10)

            add_markers(
                adata.var,
                feat,
                markers,
                ax=ax,
                cmap=cmap,
                label_col="symbol",
                legend=True,
            )

        if histtype.startswith("step"):
            # Hack to get the actual handles
            handles = [patch[0] for patch in rects]
        else:
            handles = rects
        cax.legend(
            handles=handles,
            labels=keys,
            loc="center left",
            title=fill_by,
            frameon=False,
        )
    else:
        fig, axs = configure_subplots(nplot, ncol, figsize=figsize)
        _hist_kwargs.setdefault("color", "#666666ff")

        for ax, feat in zip(axs.flat, features):
            if histtype != "line":
                n, _, rects = ax.hist(
                    adata.var[feat],
                    bins=bins,
                    log=log,
                    histtype=histtype,
                    **_hist_kwargs,
                )
            else:
                _hist_kwargs.pop("ec", None)
                _hist_kwargs.pop("edgecolor", None)
                counts, bin_edges = np.histogram(
                    adata.var[feat].dropna(inplace=False).values, bins=bins
                )
                centers = np.diff(bin_edges) / 2 + bin_edges[:-1]
                rects = ax.plot(centers, counts, **_hist_kwargs)

            add_markers(
                adata.var,
                feat,
                markers,
                ax=ax,
                cmap=cmap,
                label_col="symbol" if show_symbol else None,
                legend=True,
            )
            ax.set_xlabel(feat)

    fig.supylabel("count")

    return axs


def add_markers(
    df,
    feat,
    markers: Union[None, str, Sequence[str]],
    ax: Axes,
    cmap: Optional[str] = None,
    label_col: Optional[str] = None,
    legend: bool = True,
):
    if markers is not None:
        marker_list = [markers] if isinstance(markers, str) else markers[:]
        colors = plt.get_cmap(cmap)

        for i_marker, marker in enumerate(marker_list):
            value = df.loc[marker, feat]
            label_suffix = marker if label_col is None else df.loc[marker, label_col]
            label = f"{i_marker} - {label_suffix}"
            ax.axvline(value, color=colors(i_marker), label=label)

        ax.legend(bbox_to_anchor=(1.04, 0.5), loc="center left", title="Marker Gene")
    return ax


def plot_fitline(
    x: Union[str, Series, np.ndarray],
    y: Union[str, Series, np.ndarray],
    alternative: str = "two-sided",
    ax: Optional[Axes] = None,
    color: Optional[str] = None,
    data: Any = None,
    xlim: Optional[Tuple[Optional[float], Optional[float]]] = None,
    ylim: Optional[Tuple[Optional[float], Optional[float]]] = None,
):
    if ax is None:
        fig, ax = plt.subplots()

    if data is not None:
        xdat = data[x] if isinstance(x, str) else x[:]
        ydat = data[y] if isinstance(y, str) else y[:]
    else:
        xdat = x[:]
        ydat = y[:]

    xdat = np.array(xdat)
    ydat = np.array(ydat)

    non_nans = ~(np.isnan(xdat) | np.isnan(ydat))
    reg = linregress(xdat[non_nans], ydat[non_nans], alternative=alternative)

    xmax = xdat.max()  # max(0, xdat.max())
    xmin = xdat.min()  # min(xdat.min(), 0)
    ymin = xmin * reg.slope + reg.intercept
    ymax = xmax * reg.slope + reg.intercept

    ax.axline((xmin, ymin), (xmax, ymax), color=color)

    if xlim is not None:
        ax.set_xlim(*xlim)
    if ylim is not None:
        ax.set_ylim(*ylim)

    return ax


def scatter(
    x: str,
    y: str,
    color_by: Optional[str] = None,
    ax: Optional[Axes] = None,
    cmap: Optional[str] = None,
    legend: bool = True,
    legend_kwargs: Optional[Dict[str, Any]] = None,
    data: Optional[DataFrame] = None,
    labels: Optional[Dict[str, Any]] = None,
    rc_context: Optional[Dict[str, Any]] = None,
    fitline_kwargs: Optional[Dict[str, Any]] = None,
    contour_kwargs: Optional[Dict[str, Any]] = None,
    figsize: Optional[Tuple[float, float]] = None,
    is_categorical: Optional[bool] = None,
    aspect: Union[None, float, str] = None,
    **scatter_kwargs,
):
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = ax.get_figure()

    # cmap cannot be passed to plt.scatter if color_by is None
    if color_by is None:
        cmap = None
        scatter_kwargs.pop("vmin", None)
        scatter_kwargs.pop("vmax", None)

    _scatter_kwargs = dict(cmap=cmap, s=8)
    cmap_kwargs = dict()

    # is_categorical = False
    if data is None:
        if isinstance(color_by, str):
            raise ValueError("data must not be None if color_by is str")
        if isinstance(x, str):
            raise ValueError("data must not be None if x is str")
        if isinstance(y, str):
            raise ValueError("data must not be None if y is str")

    xdat, xstr = (data[x], x) if isinstance(x, str) else (x, None)
    ydat, ystr = (data[y], y) if isinstance(y, str) else (y, None)
    color_dat, color_str = (
        (data[color_by], color_by) if isinstance(color_by, str) else (color_by, None)
    )

    del color_by, x, y

    # if color_dat is not None and data is not None and data[color_dat].dtype == "category":
    if is_categorical or color_dat is not None and color_dat.dtype == "category":
        is_categorical = True
        _scatter_kwargs.update(
            dict(
                vmax=plt.get_cmap(cmap).N,
                c=color_dat.astype(int),
            )
        )

    elif color_dat is not None:
        is_categorical = False
        _scatter_kwargs.update(dict(c=color_dat))

    _rc_context = {}
    _rc_context.update(rc_context or {})

    with plt.rc_context(_rc_context):
        _scatter_kwargs.update(scatter_kwargs)
        scat = ax.scatter(xdat, ydat, **_scatter_kwargs)

    if contour_kwargs is not None:
        _contour_kwargs = dict(x=xdat, y=ydat, data=data)
        _contour_kwargs.update(contour_kwargs)
        contour_plot(ax, **_contour_kwargs)

    if fitline_kwargs is not None:
        _fitline_kwargs = dict(x=xdat, y=ydat, ax=ax)
        _fitline_kwargs.update(fitline_kwargs)
        ax = plot_fitline(**_fitline_kwargs)

    labels = labels or {}

    ax.set_xlabel(labels.get("x", xstr))
    ax.set_ylabel(labels.get("y", ystr))
    ax.set_title(labels.get("title", None))

    legend_kwargs = legend_kwargs or {}

    label_on_top = False
    if legend and is_categorical:
        _legend_kwargs = dict(
            bbox_to_anchor=(1.04, 0.5),
            loc="center left",
            frameon=False,
            title=color_str,
        )

        legend_kwargs.pop("label", None)
        _legend_kwargs.update(legend_kwargs)
        cax = _legend_kwargs.pop("cax", None)
        if cax is None:
            cax = ax
        else:
            _legend_kwargs.pop("bbox_to_anchor", None)
        cax.legend(*scat.legend_elements(), **_legend_kwargs)

    elif legend and color_dat is not None:
        # I'm not 100% certain which one to use
        use_divider = not True

        if use_divider:
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad="2%")
            cmap_kwargs["cax"] = cax
        else:
            cax = None
            cmap_kwargs["ax"] = ax
            cmap_kwargs["use_gridspec"] = True

        cmap_kwargs.update(legend_kwargs or {})
        cbar_title = cmap_kwargs.pop("title", None)

        cbar = fig.colorbar(scat, **cmap_kwargs)

        if cbar_title is not None and label_on_top:
            cbar.ax.set_title(cbar_title, fontsize=8)
        elif cbar_title is not None:
            cbar.set_label(cbar_title)

    if aspect is not None:
        ax.set_aspect(aspect)
    return ax



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from voyagerpy.utils.utils import (
    add_per_cell_qcmetrics,
    add_per_gene_qcmetrics,
    get_scale,
    is_highres,
    kurtosis,
    listify,
    log_norm_counts,
    make_unique,
    normalize_csr,
    scale,
)

from voyagerpy.utils.hvg import get_top_hvgs, model_gene_var
from voyagerpy.utils.markers import get_marker_genes, get_p_clusters, find_markers, get_stat_clusters

__all__ = [
    "add_per_cell_qcmetrics",
    "add_per_gene_qcmetrics",
    "find_markers",
    "get_marker_genes",
    "get_p_clusters",
    "get_top_hvgs",
    "get_scale",
    "get_stat_clusters",
    "is_highres",
    "kurtosis",
    "listify",
    "log_norm_counts",
    "make_unique",
    "model_gene_var",
    "normalize_csr",
    "scale",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import libpysal
# need morean for data on obs, spatial and non spatial.
# then need moran for 
def moran(adata, gene):
    from esda.moran import Moran
    # get graph object
    wg = get_visium_graph(adata)
    # y = get gene data from adata.X.todense()

    y = get_correct_gene_form_from_genes(gene)
    mi = Moran(y, wg)
    mi.I


function_map = {
  'moran': moran,
  'geary': calcEngagement,
  'moran.mc':,
  'geary.mc',
  'moran.test',
  'geary.test',
  'globalG.test',
  'sp.correlogram',
  'moran.plot',
  'localmoran',
  'localmoran_perm',
  'localC',
  'localC_perm',
  'localG',
  'localG_perm',
  'LOSH',
  'LOSH.mc',
  'LOSH.cs'
  
}

def calculate_univariate(
        adata,
        fun_type,
        features,
        barcode_graph_name,
        zero_policy=None,
        in_place=False,
        include_self=None,
        p_adjust_method=None
):
    if(fun_type in function_map):
        val = function_map[fun_type](adata, features)

    return val






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from anndata import AnnData
import numpy as np
from numpy import ndarray
import pandas as pd
from pandas import DataFrame
from scipy.stats import mannwhitneyu
from scipy import sparse
from typing import Union


def get_p_vals(
    adata: AnnData, clust1: str, clust2=None, test: str = "mw", alternative: str = "greater", skip=None, clust_str: str = "cluster"
):
    """Run a given statistical test for genes in a given anndata object,between cells in clust1 and clust2.

    Parameters
    ----------
    adata : anndata object
        Anndata object with that has already been given a leiden clustering, or custom one in adata.obs["cluster"]
    clust1 : String
        A single group of cells, as defined by the clustering.
    clust2 : String
        String defining the second group that is to be tested against the first,
        if this is none then the the test is cluster1 vs rest. The default is None.
    test : String, optional
        Type of test. For now only mann whitney is possible. The default is "mw".
    alternative : String, optional
        Direction of test. The default is "greater".
    skip : List or array of ints, optional
        This is a parameter to speed up calculations. The default method for combining p values,
        takes the largest one for each cluster, therefore 1 has already been seen in the row, that row is skipped,
        because that row will always be designated a p value of 1.
        The default is None.

    Returns
    -------
    arr_final : Array of shape n where n is adata.shape[1]
        The p values for each gene.

    """
    # data_is_AnnData = isinstance(data, AnnData)
    # if data_is_AnnData:

    if not sparse.issparse(adata.X):
        adata.X = sparse.csr_matrix(adata.X)

    # if "highly_variable" in adata.var:
    #    genes = np.array(adata.var.index.get_indexer(adata.var.loc[adata.var["highly_variable"] == True].index))
    # else:
    # genes = np.array(adata.var.index.get_indexer(adata.var.index))
    cl_index = adata.obs.index.get_indexer(adata.obs.loc[adata.obs[clust_str] == clust1].index)
    arr_final = np.ones((adata.shape[1]), dtype=float)
    if clust2 is None:
        comp_index = adata.obs.index.get_indexer(adata.obs.loc[adata.obs[clust_str] != clust1].index)

    else:
        comp_index = adata.obs.index.get_indexer(adata.obs.loc[adata.obs[clust_str] == clust2].index)
    X_cl = adata.X.todense()[cl_index, :]
    X_comp = adata.X.todense()[comp_index, :]

    if skip is not None:
        it_values = np.delete(np.array(range(adata.shape[1])), skip)
        X_cl = adata.X.todense()[cl_index[:, None], it_values]
        X_comp = adata.X.todense()[comp_index[:, None], it_values]
    else:
        it_values = range(adata.shape[1])
    if test == "mw":
        mw = mannwhitneyu(X_cl, X_comp, alternative=alternative)
        arr = mw[1]
    else:
        raise NotImplementedError(f"Test '{test}' is not implemented")

    if skip is not None:
        arr_final[it_values] = arr

    else:
        arr_final = arr

    return arr_final


def get_statistics(adata: AnnData, clust1: str, clust2=None, test: str = "mw", alternative: str = "greater", clust_str: str = "cluster"):
    """\
    Get summary statistics for genes.

    Similar to get_p_vals function,
    except this function returns additional info on effect size and the test statistic.


    Parameters
    ----------
    adata : Anndata object
        Anndata object with clustering having been defined in adata.obs.
    clust1 : String
        String of a single given cluster to identify rows(cells) in the cluster in adata.obs.
    clust2 : String, optional
        Group to test per gene vs clust1, if none then a clust1 vs rest test is performed. The default is None.
    test : String, optional
        The given statistical test to perform, per gene. The default is "mw".
    alternative : String, optional
        Direction/type of test. The default is "greater".
    clust_str : str, optional
        The string to identify the cluster column in adata.obs. The default is "cluster".

    Raises
    ------
    NotImplementedError
        In future more tests will be implemented,
        for now raise and error if a test is called that does not exist.

    Returns
    -------
    p_vals : np.ndarray
        p values for the test, each row is a different gene.
    test_stat : np.ndarray
        test statistic for each test.
    effect_size : np.ndarray
        The effect size for each test. In the case of the Mann-Whitney test it is defined as the AUC.
    """
    if not sparse.issparse(adata.X):
        adata.X = sparse.csr_matrix(adata.X)

    cl_index = adata.obs.index.get_indexer(adata.obs.loc[adata.obs[clust_str] == clust1].index)
    if clust2 is None:
        comp_index = adata.obs.index.get_indexer(adata.obs.loc[adata.obs[clust_str] != clust1].index)

    else:
        comp_index = adata.obs.index.get_indexer(adata.obs.loc[adata.obs[clust_str] == clust2].index)
    X_cl = adata.X.todense()[cl_index, :]
    X_comp = adata.X.todense()[comp_index, :]

    if test == "mw":
        mw = mannwhitneyu(X_cl, X_comp, alternative=alternative)
        test_stat = mw[0]
        p_vals = mw[1]
        effect_size = test_stat / (X_cl.shape[0] * X_comp.shape[0])

    else:
        raise NotImplementedError(f"Test '{test}' is not implemented")

    return p_vals, test_stat, effect_size


def get_p_clusters(
    adata: AnnData, clust: Union[str, int], skip_precalc: bool = False, pval_type: str = "all", cluster: str = "cluster"
) -> Union[DataFrame, ndarray]:
    # rows = adata.X.shape[0]
    pval_arrs = []
    # nr_clust = adata.obs["cluster"].cat.codes
    colnames = []
    if skip_precalc:
        first = False
        skip_index = None
    else:
        first = True
        skip_index = None
    for i in np.array(adata.obs[cluster].cat.categories):
        if str(clust) == str(i):
            continue
        arr = get_p_vals(adata, str(clust), str(i), test="mw", skip=skip_index, clust_str=cluster)
        if first:
            skip_index = np.where(arr == 1)[0]
            first = False
        pval_arrs.append(arr)
        colnames.append("clust" + str(clust) + "vs" + str(i))

    pval_np = np.array(pval_arrs)

    pval_np = np.transpose(pval_np.reshape((pval_np.shape[0], pval_np.shape[1])))

    df = pd.DataFrame(pval_np, index=adata.var.index, columns=colnames)

    if pval_type == "all":
        arr = df.max(axis=1)
    else:
        arr = df

    return arr


def get_stat_clusters(adata: AnnData, clust: Union[str, int], cluster: str = "cluster") -> dict:
    """\
    Get differential expression statistics on genes for a given cluster, against all other clusters.

    Parameters
    ----------
    adata : AnnData
        AnnData object with defined clusters in adata.obs.
    clust : Union[str, int]
        Perform tests on genes, expressed in cells in this cluster.
    cluster : str, optional
        String to identify cluster column in adata.obs. The default is "cluster".

    Returns
    -------
    dict
        Dictionary with 3 entries, p values, test statistics and effect sizes,
        defined for each gene and each cluster paired with the one in the clust parameter.

    """
    pvals = []
    test_stat = []
    es = []
    colnames = []
    for i in np.array(adata.obs[cluster].cat.categories):
        if str(clust) == str(i):
            continue
        p_arr, mw_arr, es_arr = get_statistics(adata, str(clust), str(i), test="mw", clust_str=cluster)
        pvals.append(p_arr)
        test_stat.append(mw_arr)
        es.append(es_arr)
        colnames.append("clust" + str(clust) + "vs" + str(i))
    val_dict = {"p_values": pvals, "test_statistic": test_stat, "effect_size": es}
    for k, v in val_dict.items():
        temp_np = np.array(v)
        temp_np = np.transpose(temp_np.reshape((temp_np.shape[0], temp_np.shape[1])))
        temp_df = pd.DataFrame(temp_np, index=adata.var.index, columns=colnames)
        val_dict[k] = temp_df
    return val_dict


def get_marker_genes(adata: AnnData, hvg: bool = False, hvg_string: str = "highly_variable", cluster: str = "cluster") -> DataFrame:
    """\
    Find marker genes for each cluster.

    Similar to find_markers but returns only,
    top marker genes for each cluster, with no additional information.

    Parameters
    ----------
    adata : AnnData
        AnnData object with defined clustering in adata.obs.
    hvg : bool, optional
        Whether to only test highly variable genes. The default is False.
    hvg_string : str, optional
        String to identify highly variable genes in adata.var. The default is "highly_variable".
    cluster : str, optional
        String to identify cluster column in adata.obs. The default is "cluster".

    Returns
    -------
    DataFrame
        DataFrame with each column corresponding to differentially
        expressed genes for each cluster in descending order.

    """
    if cluster not in adata.obs:
        Exception("There need to be clusters defined for this anndata object")
    nr_clust = adata.obs[cluster].cat.codes
    markers_desc = []
    adata_obj = adata if not hvg else adata[:, adata.var[hvg_string]].copy()
    for i in np.array(adata.obs[cluster].cat.categories):
        curr_cl_data = get_p_clusters(adata_obj, i, cluster=cluster)
        markers_desc.append(np.array(curr_cl_data[np.argsort(curr_cl_data)].index))
    markers = pd.DataFrame(np.transpose(markers_desc), columns=[f"{cluster}_{i}" for i in range(np.max(nr_clust) + 1)])

    return markers


def find_markers(adata: AnnData, hvg: bool = False, hvg_string: str = "highly_variable", cluster: str = "cluster") -> dict:
    """\
    Return summary statistics on differentially expressed genes by groups defined in cluster.

    This function is based on findMarkers in scran.

    Parameters
    ----------
    adata : AnnData
        AnnData object with defined clustering in adata.obs.
    hvg : bool, optional
        Whether to only test highly variable genes. The default is False.
    hvg_string : str, optional
        String to identify highly variable genes in adata.var. The default is "highly_variable".
    cluster : str, optional
        String to identify cluster column in adata.obs. The default is "cluster".

    Returns
    -------
    dict
        Returns dictionary with summary statistics.
        P values, Benjamini-Hochberg False discovery rate,
        statistical measure of effect size, default for Mann-Whitney is AUC.

    """
    if cluster not in adata.obs:
        Exception("There need to be clusters defined for this anndata object")
    adata_obj = adata if not hvg else adata[:, adata.var[hvg_string]].copy()
    res = {}
    for i in np.array(adata.obs[cluster].cat.categories):
        stat_dict = get_stat_clusters(adata_obj, i, cluster=cluster)
        pval_max = stat_dict["p_values"].max(axis=1)
        index_of_maxs = stat_dict["p_values"].columns.get_indexer(stat_dict["p_values"].idxmax(axis=1))
        stat_dict["effect_size"]
        sort_order = np.argsort(pval_max)
        es = np.array(stat_dict["effect_size"])[np.arange(len(stat_dict["effect_size"])), index_of_maxs]
        res[f"{cluster}_{i}"] = pd.DataFrame(
            {"p_vals": pval_max[sort_order], "FDR": p_adjust_bh(pval_max)[sort_order], "summary_es": es[sort_order]},
            index=np.array(pval_max[sort_order].index, dtype=str),
        )

    return res


# taken from https://stackoverflow.com/questions/7450957/how-to-implement-rs-p-adjust-in-python/33532498#33532498
def p_adjust_bh(p):
    """Benjamini-Hochberg p-value correction for multiple hypothesis testing."""
    p = np.asfarray(p)
    by_descend = p.argsort()[::-1]
    by_orig = by_descend.argsort()
    steps = float(len(p)) / np.arange(len(p), 0, -1)
    q = np.minimum(1, np.minimum.accumulate(steps * p[by_descend]))
    return q[by_orig]



from typing import Tuple

import numpy as np
from matplotlib import colormaps
from matplotlib.colors import LinearSegmentedColormap, ListedColormap, Normalize


def register_segmented_cmap(cmap_name: str, cmap_colors: Tuple[str], reverse=False) -> None:
    if cmap_name in colormaps:
        return

    if reverse:
        cmap_colors = cmap_colors[::-1]

    # cmap = LinearSegmentedColormap.from_list(cmap_name, cmap_colors, N=len(cmap_colors))
    cmap = LinearSegmentedColormap.from_list(cmap_name, cmap_colors)
    colormaps.register(cmap)
    colormaps.register(cmap.reversed())


def register_listed_cmap(cmap_name: str, cmap_colors: Tuple[str], reverse: bool = False) -> None:
    if cmap_name in colormaps:
        return

    if reverse:
        cmap_colors = cmap_colors[::-1]

    dittoseq_cmap = ListedColormap(cmap_colors, name=cmap_name, N=len(cmap_colors))
    colormaps.register(dittoseq_cmap)
    colormaps.register(dittoseq_cmap.reversed())


class DivergentNorm(Normalize):
    def __init__(self, vmin, vmax, vcenter=0, clip=False):
        super().__init__(vmin=vmin, vmax=vmax, clip=clip)

        # The largest interval defines the slope
        dx = max(abs(vmax - vcenter), abs(vcenter - vmin))

        # The y difference of the dx is always 0.5
        dy = 0.5
        slope = dy / dx
        intercept = 0.5 - slope * vcenter

        f = lambda x: slope * x + intercept
        self.x = [self.vmin, self.vmax]
        self.y = [f(self.vmin), f(self.vmax)]

    def __call__(self, value, clip=None):
        return np.ma.masked_array(np.interp(value, self.x, self.y))

    def inverse(self, value):
        return np.ma.masked_array(np.interp(value, [0, 1], self.x, left=self.vmin, right=self.vmax))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from voyagerpy.spatial.graphs import find_visium_graph
from voyagerpy.spatial.spatial import (
    apply_transforms,
    cancel_transforms,
    compute_correlogram,
    compute_higher_order_neighbors,
    compute_spatial_lag,
    detect_tissue_threshold,
    get_approx_tissue_boundary,
    get_default_graph,
    get_geom,
    get_spot_coords,
    get_tissue_boundary,
    get_tissue_contour_score,
    local_moran,
    losh,
    mirror_img,
    moran,
    rollback_transforms,
    rotate_img90,
    set_default_graph,
    set_geometry,
    to_points,
    to_spatial_weights,
)

__all__ = [
    "apply_transforms",
    "cancel_transforms",
    "compute_correlogram",
    "compute_higher_order_neighbors",
    "compute_spatial_lag",
    "detect_tissue_threshold",
    "get_approx_tissue_boundary",
    "get_default_graph",
    "get_geom",
    "get_spot_coords",
    "get_tissue_boundary",
    "get_tissue_contour_score",
    "local_moran",
    "losh",
    "mirror_img",
    "moran",
    "rollback_transforms",
    "rotate_img90",
    "set_default_graph",
    "set_geometry",
    "to_points",
    "to_spatial_weights",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from math import exp

import numpy as np
import pandas as pd

from scipy.optimize import least_squares
from scipy import sparse
from scipy.stats import iqr, norm
from sklearn.linear_model import LinearRegression

from statsmodels.stats.multitest import fdrcorrection
from statsmodels.nonparametric.kde import KDEUnivariate
from scanpy.preprocessing._utils import _get_mean_var


def get_parametric_start(_means, _vars, left_n=100, left_prop=0.1, grid_length=10, b_grid_range=5, n_grid_max=10):
    """Get initial guess of parameters a,b,n in for for nonlinear optimization of function f.

    Parameters
    ----------
    _means : array like of length n
        Array of means
    _vars : array like of length n
        Array of variances
    left_n : int, optional
        How many starting points to use to estimate a. The default is 100.
    left_prop : float, optional
        Proporiton of points to use to estimate a. The default is 0.1.
    grid_length : float, optional
        length of grid for . The default is 10.
    b_grid_range : float, optional
        range fro b parameter grid search. The default is 5.
    n_grid_max : float, optional
        max for n parameter grid search. The default is 10.

    Returns
    -------
    a_start : float
        Initial guess for parameter a for minimization
    b_start : float
        Initial guess for parameter b for minimization
    n_start : float
        Initial guess for parameter b for minimization

    """
    o = np.sort(_means)
    n = _vars.shape[0]

    left_n = min(left_n, n * left_prop)
    keep = pd.Index(_means).get_indexer(o[: max(1, 100)])

    y = _vars[keep]

    x = _means[keep]
    lm = LinearRegression(fit_intercept=False)
    grad = lm.fit(x.reshape(-1, 1), y).coef_[0]

    b_grid_pts = 2 ** np.linspace(-b_grid_range, b_grid_range, n_grid_max)
    n_grid_pts = 2 ** np.linspace(0, n_grid_max, grid_length)
    hits = np.array([(x, y) for x in b_grid_pts for y in n_grid_pts], dtype=np.float64)

    possible_vals = np.apply_along_axis(lambda x: sum((_vars - (1.13 * x[0] * _means) / ((_means ** x[1]) + x[0] + 0.001)) ** 2), 1, hits)

    min_ind = np.argmin(possible_vals)
    b_start = np.log(hits[min_ind][0])
    n_start = np.log(max(1e-8, hits[min_ind][1] - 1))
    a_start = np.log(grad * hits[min_ind][0])

    return a_start, b_start, n_start


def inverse_density_weights(x, adjust=1):
    """\
    Calculate inverse density weights for data x.

    Parameters
    ----------
    x : ndarray
        Array of points, in this case means.
    adjust : int, optional
        DESCRIPTION. The default is 1.

    Raises
    ------
    Exception
        Fail if zero length of means.

    Returns
    -------
    w : ndarray
        Array of inverse density weights.

    """
    if len(x) == 0:
        raise Exception("Cannot parse nonzero array")
    kde = KDEUnivariate(x)
    bw = 0.9 * min(np.std(x), iqr(x) / 1.34) * (len(x) ** (-1 / 5))
    # out = kde.fit(bw=bw,gridsize=512,clip=(min(m),max(m)),cut=0)
    out = kde.fit(bw=bw, gridsize=512, cut=0)
    kde_supp = out.support
    # kde_dens = kde.density
    kde_out = out.evaluate(kde_supp)

    w = 1 / np.interp(x, kde_supp, kde_out)
    w = w / np.mean(w)
    return w


# define your function f with inputs x and parameters a, b, c
def f(x, a, b, n):
    """\
    Variance function.

    This function is used to describe the variances as a function of means.

    Parameters
    ----------
    x : ndarray
        Values to calculate
    a : float
        Parameter a.
    b : float
        Parameter b.
    n : float
        Parameter n.

    Returns
    -------
    TYPE
        return f(x).

    """
    return (exp(a) * x) / (x ** (1 + exp(n)) + exp(b))


# define the residual function with weights, which is the difference between the function and the data


def residual(params, _mean, _vars, _weight):
    """\
    Pseudo residual function for f.

    Implemented with weights

    Parameters
    ----------
    params : List
        Params for function f.
    _mean : ndarray
        Array of means.
    _vars : ndarray
        Array of variances.
    _weight : ndarray
        Array of weights.

    Returns
    -------
    rss : ndarray
        weighted residuals.

    """
    a, b, n = params

    rss = (_vars - f(_mean, a, b, n)) * np.sqrt(_weight)

    return rss


def parametric_fit(_mean, _vars, _weight, start):
    """\
    Fit function with parameters for given data.

    Parameters
    ----------
    _mean : ndarray
        Array of means.
    _vars : TYPE
        Array of variances.
    _weight : TYPE
        Array of weights.
    start : TYPE
        Initial start values of parameters a,b,n.

    Returns
    -------
    a : TYPE
        Best guess of a.
    b : TYPE
        Best guess of b.
    n : TYPE
        Best guess of n.

    """
    x = _mean
    y = _vars
    w = _weight  # set weights for each point

    # set initial guess for the parameters
    params0 = np.array(start)
    # run the optimization
    res = least_squares(residual, params0, args=(x, y, w), ftol=1e-12, verbose=0, max_nfev=1200)  # ,method="dogbox")

    # get the optimized parameters
    a, b, n = res.x

    return a, b, n


def f_predict(x, a, b, n):
    a0 = exp(a)
    b0 = exp(b)
    n0 = exp(n) + 1
    return (a0 * x) / (x**n0 + b0)


def correct_logged_expectation(x, y, w, FUN):
    leftovers = y / FUN(x)
    med = weighted_median(leftovers, w)
    OUT = lambda x: FUN(x) * med

    std_dev = weighted_median(abs((leftovers / med) - 1), w) * 1.4826

    return OUT, std_dev


def weighted_median(x, w):
    """\
    Return weighted median of x.

    Parameters
    ----------
    x : ndarray
        data points.
    w : ndarray
        weights.

    Returns
    -------
    float
        weighted median of x.

    """
    if x.shape != w.shape or w is None:
        w = np.ones(x.shape[0])

    # o = pd.Index(x).get_indexer(np.sort(x))
    o = np.argsort(x)

    x = x[o]
    w = w[o]
    p = np.cumsum(w) / sum(w)
    n = np.where(p < 0.5)[0].shape[0]
    if p[n] < 0.5:
        return x[n]
    else:
        return (x[n] + x[n + 1]) / 2

    # return n


def decompose_log_exprs(_means, _vars, fit_means, fit_vars, names=None):
    """\
    Decompose the variance into technical and biological.

    Parameters
    ----------
    _means : ndarray
        Means.
    _vars : ndarray
        Variances.
    fit_means : ndarray
        Kept for compatibility in R.
    fit_vars : ndarray
        Kept for compatibility in R.
    names : array-like, optional
        Gene names. The default is None.

    Returns
    -------
    output : DataFrame
        Returns dataframe with relevant columns.

    """
    fit, std_dev = fit_trend_var(fit_means, fit_vars)

    output = pd.DataFrame({"mean": _means, "total": _vars, "tech": fit(_means)}, index=names)
    output["bio"] = output["total"] - output["tech"]
    output["p_value"] = 1 - norm.cdf(output["bio"] / output["tech"], scale=std_dev)
    filt_out = output[~output["p_value"].isna()]
    # filt_out["FDR"] = fdrcorrection(filt_out["p_value"])[1]
    _, pval_corr = fdrcorrection(filt_out["p_value"])
    output.loc[filt_out.index, "FDR"] = pval_corr

    return output
    # 1 - norm().cdf(0.1)


def model_gene_var(adata, block=None, design=None, subset_row=None, subset_fit=None, gene_names=None):
    """\
    Return the modelled gene variance.

    Modelled on similar method in SCRAN package.
    Parameters
    ----------
    adata : AnnData
        DESCRIPTION.
    block : None, optional
        Compatibility with R. The default is None.
    design : None, optional
        Compatibility with R. The default is None.
    subset_row : ndarray, optional
        If only a subset of the rows are used. The default is None.
    subset_fit : bool, optional
        if subsetting should occur, yet to be implmented. The default is None.
    gene_names : ndarray, optional
        If gene names are provided. The default is None.

    Returns
    -------
    collected : DataFrame
        Information on modelled variance into biological and technical.

    """
    if sparse.issparse(adata):
        x_means, x_vars = _get_mean_var(adata, axis=0)
        names = gene_names
    else:
        x_means, x_vars = _get_mean_var(adata.X, axis=0)
        names = adata.var.index
    # decompose_log_exprs(x_means, x_vars)
    if subset_fit is None:
        fit_stats_means = x_means
        fit_stats_vars = x_vars
    else:
        return None

    collected = decompose_log_exprs(x_means, x_vars, fit_stats_means, fit_stats_vars, names=names)

    return collected


def fit_trend_var(gene_mean, gene_var, min_mean=0.1, parametric=True, lowess=False, density_weights=True):
    """\
    Fit the variance function to the data.

    Parameters
    ----------
    gene_mean : ndarray
        mean of each gene.
    gene_var : ndarray
        variances for each gene.
    min_mean : float, optional
        Minimum mean for a gene to be included in fitting the trend. The default is 0.1.
    parametric : bool, optional
        Whether to run a parametric trend fit. The default is True.
    lowess : bool, optional
        Whether to run a lowess trend fit. The default is False.
    density_weights : bool, optional
        Whether to include density weights in the fitting. The default is True.

    Returns
    -------
    fit : function
        fitted and corrected variance function.
    std_dev : TYPE
        standard deviation of fit.

    """
    is_okay = np.intersect1d(np.where(gene_var > 1.0e-8), np.where(gene_mean >= min_mean))
    v = gene_var[is_okay]
    m = gene_mean[is_okay]
    if density_weights:
        w = inverse_density_weights(m)
    else:
        w = np.ones(len(m))

    to_fit = np.log(v)

    left_edge = min(m)
    # PARAMFUN = lambda x: np.min(np.column_stack((np.ones(len(x)), x / left_edge)), axis=1)
    PARAMFUN = lambda x: np.minimum(x / left_edge, 1)

    if parametric:
        a_start, b_start, n_start = get_parametric_start(m, v)
        a, b, n = parametric_fit(m, v, w, [a_start, b_start, n_start])
        # to_fit = to_fit - np.log(f(m, a, b, n))
        PARAMFUN = lambda x: f_predict(x, a, b, n)
        # return PARAMFUN

    if lowess:
        # idx = np.round(np.linspace(0, len(m) - 1, 200)).astype(int)
        # ll = lowess(to_fit, exog=m, xvals=idx, resid_weights=w)
        pass
    else:
        UNSCALEDFUN = PARAMFUN
    fit, std_dev = correct_logged_expectation(m, v, w, UNSCALEDFUN)
    return fit, std_dev


def get_top_hvgs(stats, n=2000, stat="bio", var_threshold=0):
    return np.array(stats.nlargest(n, "bio").index, dtype="str")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from voyagerpy import metrics as mtr
from voyagerpy import plotting as plt
from voyagerpy import spatial as spt
from voyagerpy import utils as utl
from voyagerpy.read import read_10x_visium
from voyagerpy.read import read_10x_counts

from anndata import AnnData
from geopandas import GeoDataFrame
from pandas import Index

from typing import Union, Tuple, TypeVar
from scipy import sparse
import numpy as np

from anndata import AnnData
import anndata


class Voyager(AnnData):
    def __init__(self, *args, **kwargs):
        X = args[0] if len(args) else kwargs.get("X", None)

        dtype_present = len(args) > 8 or "dtype" in kwargs
        dtype = None

        if not dtype_present:
            if isinstance(X, (AnnData, Voyager)):
                dtype = X.X.dtype
            elif X is None:
                dtype = None
            else:
                dtype = X.dtype

        _kwargs = dict(dtype=dtype)
        _kwargs.update(kwargs)

        super().__init__(*args, **_kwargs)
        barcode_geo = self.obsm.setdefault("geometry", gpd.GeoDataFrame(index=self.obs_names))

        if not isinstance(barcode_geo, anndata._core.views.DataFrameView) and not isinstance(barcode_geo, gpd.GeoDataFrame):
            self.obsm["geometry"] = gpd.GeoDataFrame(barcode_geo)

    def copy(self, *args) -> "Voyager":
        adata = super().copy(*args)
        return Voyager(adata)

    def __getitem__(self, index):
        oidx, vidx = self._normalize_indices(index)
        return Voyager(self, oidx=oidx, vidx=vidx, asview=True)


__all__ = [
    "metrics",
    "plotting",
    "spatial",
    "utils",
    "read_10x_counts",
    "read_10x_visium",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from matplotlib.pyplot import rcParams

from voyagerpy.plotting.cmap_helper import register_listed_cmap, register_segmented_cmap
from voyagerpy.plotting.plot import (
    elbow_plot,
    grouped_violinplot,
    imshow,
    moran_plot,
    plot_barcode_data,
    plot_barcode_histogram,
    plot_barcodes_bin2d,
    plot_bin2d,
    plot_correlogram,
    plot_dim_loadings,
    plot_expression,
    plot_features_bin2d,
    plot_features_histogram,
    plot_fitline,
    plot_local_result,
    plot_moran_mc,
    plot_pca,
    plot_spatial_feature,
    scatter,
    simple_violinplot,
    spatial_reduced_dim,
)

# https://github.com/dtm2451/dittoSeq
from .ditto_colors import ditto_colors

# https://github.com/thomasp85/scico
from .roma_colors import roma_colors

# https://github.com/cran/RColorBrewer
from .blues_colors import blues3, blues4, blues5, blues6, blues7, blues8, blues9


def set_default_cmap(cmap_name: str) -> None:
    rcParams["image.cmap"] = cmap_name


register_segmented_cmap("roma", roma_colors, reverse=True)  # type: ignore
register_segmented_cmap("Blues3", blues3, reverse=False)  # type: ignore
register_segmented_cmap("Blues4", blues4, reverse=False)  # type: ignore
register_segmented_cmap("Blues5", blues5, reverse=False)  # type: ignore
register_segmented_cmap("Blues6", blues6, reverse=False)  # type: ignore
register_segmented_cmap("Blues7", blues7, reverse=False)  # type: ignore
register_segmented_cmap("Blues8", blues8, reverse=False)  # type: ignore
register_segmented_cmap("Blues9", blues9, reverse=False)  # type: ignore


register_listed_cmap("dittoseq", ditto_colors, reverse=False) # type: ignore
set_default_cmap("dittoseq")

__all__ = [
    "elbow_plot",
    "grouped_violinplot",
    "imshow",
    "moran_plot",
    "plot_barcode_data",
    "plot_barcode_histogram",
    "plot_barcodes_bin2d",
    "plot_bin2d",
    "plot_correlogram",
    "plot_dim_loadings",
    "plot_expression",
    "plot_features_bin2d",
    "plot_features_histogram",
    "plot_pca",
    "plot_spatial_feature",
    "simple_violinplot",
    "spatial_reduced_dim",
    "scatter",
    "plot_fitline",
    "plot_local_result",
    "plot_moran_mc",
]
